//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.core.dom;

import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ClassInstanceCreation;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.ExpressionStatement;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.SimpleType;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;

/**
 * @author Yoonki Song
 */
@SuppressWarnings("unchecked")
public final class ASTNodeFactory {

	/**
	 * no instance.
	 */
	private ASTNodeFactory() {
	}

	/**
	 * Creates a VariableDecleartionStatement.
	 * <p>
	 * Example.
	 * 
	 * <pre>
	 * String str = builder.toString();
	 * </pre>
	 * 
	 * @param ast
	 *            the AST.
	 * @param typeId
	 *            the type identifier.
	 * @param varId
	 *            the variable identifier.
	 * @param objId
	 *            the object identifier.
	 * @param methodId
	 *            the method identifier.
	 * @return the new variable declaration statement.
	 */
	public static VariableDeclarationStatement createVariableDeclarationStatement2(final AST ast, final String typeId,
			final String varId, final String objId, final String methodId) {
		VariableDeclarationStatement variableDeclarationStatement = (VariableDeclarationStatement) ast
				.createInstance(VariableDeclarationStatement.class);
		variableDeclarationStatement.setType(createSimpleType(ast, typeId));
		variableDeclarationStatement.fragments().add(createVariableDeclarationFragment2(ast, varId, objId, methodId));
		return variableDeclarationStatement;
	}

	/**
	 * Creates an ExpressionStatement with MethodInvocation.
	 * 
	 * <p>
	 * Example.
	 * 
	 * <pre>
	 * builder.toString();
	 * </pre>
	 * 
	 * @param ast
	 *            the AST
	 * @param varId
	 *            the variable identifier
	 * @param methodId
	 *            the method identifier
	 * @param rhs
	 *            the right hand side.
	 * @return the new ExpressionStatement.
	 */
	public static ExpressionStatement createExpressionStatementWithMI(final AST ast, final String varId, final String methodId,
			final Expression rhs) {
		return ast.newExpressionStatement(createMethodInvocation(ast, varId, methodId, rhs));
	}

	/**
	 * Creates a method invocation (Expression).
	 * 
	 * <p>
	 * Example.
	 * 
	 * <pre>
	 * builder.toString()
	 * </pre>
	 * 
	 * @param ast
	 *            the AST.
	 * @param varIdentifer
	 *            the variable identifier.
	 * @param methodIdentifier
	 *            the method identifier.
	 * @param rhs
	 *            the right hand side expression.
	 * @return the new method invocation.
	 */
	public static MethodInvocation createMethodInvocation(final AST ast, final String varIdentifer,
			final String methodIdentifier, final Expression rhs) {
		MethodInvocation methodInvocation = ast.newMethodInvocation();
		methodInvocation.setExpression(createSimpleName(ast, varIdentifer));
		methodInvocation.setName(createSimpleName(ast, methodIdentifier));
		methodInvocation.arguments().add(rhs);
		return methodInvocation;
	}

	/**
	 * Creates a SimpleType.
	 * 
	 * @param ast
	 *            the AST.
	 * @param identifier
	 *            the identifier.
	 * @return new SimpleType.
	 */
	public static SimpleType createSimpleType(final AST ast, final String identifier) {
		SimpleType simpleType = ast.newSimpleType(createSimpleName(ast, identifier));
		return simpleType;
	}

	/**
	 * Creates a SimpleName.
	 * 
	 * @param ast
	 *            the AST.
	 * @param identifier
	 *            the identifier.
	 * @return new SimpleName.
	 */
	public static SimpleName createSimpleName(final AST ast, final String identifier) {
		SimpleName simpleName = ast.newSimpleName(identifier);
		return simpleName;
	}

	/**
	 * Creates a VariableDeclarationFragment.
	 * 
	 * <pre>
	 * str = new String(&quot;&quot;);
	 * </pre>
	 * 
	 * @param ast
	 *            the AST.
	 * @param typeIdentifier
	 *            the type identifier
	 * @param varIdentifier
	 *            the variable identifier
	 * @return new VariableDeclarationFragment
	 */
	public static VariableDeclarationFragment createVariableDeclarationFragment(final AST ast, final String typeIdentifier,
			final String varIdentifier) {
		VariableDeclarationFragment variableDeclarationFragment = ast.newVariableDeclarationFragment();
		variableDeclarationFragment.setName(createSimpleName(ast, varIdentifier));
		variableDeclarationFragment.setExtraDimensions(0);
		variableDeclarationFragment.setInitializer(createClassInstanceCreation(ast, typeIdentifier));
		return variableDeclarationFragment;
	}

	/**
	 * Creates a ClassInstanceCreation.
	 * <p>
	 * Example
	 * 
	 * <pre>
	 * </pre>
	 * 
	 * @param ast
	 *            the AST.
	 * @param typeIdentifier
	 *            the type identifier
	 * @return new ClassInstanceCreation
	 */
	public static ClassInstanceCreation createClassInstanceCreation(final AST ast, final String typeIdentifier) {
		ClassInstanceCreation classInstanceCreation = ast.newClassInstanceCreation();
		classInstanceCreation.setExpression(null);
		classInstanceCreation.setType(createSimpleType(ast, typeIdentifier));
		classInstanceCreation.setAnonymousClassDeclaration(null);
		return classInstanceCreation;
	}

	// initialized by MethodInvocation
	/**
	 * 
	 * @param ast
	 *            the AST.
	 * @param varId
	 *            the variable identifier.
	 * @param objId
	 *            the object identifier.
	 * @param methodId
	 *            the method identifier.
	 * @return new VariableDeclarationFragment.
	 */
	public static VariableDeclarationFragment createVariableDeclarationFragment2(final AST ast, final String varId,
			final String objId, final String methodId) {
		VariableDeclarationFragment variableDeclarationFragment = ast.newVariableDeclarationFragment();
		variableDeclarationFragment.setName(createSimpleName(ast, varId));
		variableDeclarationFragment.setExtraDimensions(0);
		variableDeclarationFragment.setInitializer(createtMethodInvocation(ast, objId, methodId));
		return variableDeclarationFragment;
	}

	/**
	 * Creates a MethodInvocation.
	 * 
	 * <pre>
	 * 
	 * </pre>
	 * 
	 * @param ast
	 *            the AST.
	 * @param objId
	 *            the object identifier.
	 * @param methodId
	 *            the method identifier.
	 * @return new method invocation.
	 */
	public static MethodInvocation createtMethodInvocation(final AST ast, final String objId, final String methodId) {
		MethodInvocation methodInvocation = ast.newMethodInvocation();
		methodInvocation.setExpression(createSimpleName(ast, objId));
		methodInvocation.setName(createSimpleName(ast, methodId));
		return methodInvocation;
	}

	// /**
	// * Creates a new parser which can be used to create ASTs for compilation
	// * units.
	// *
	// * @param compilationUnit
	// * the compilation unit
	// * @param monitor
	// * the progress monitor
	// * @return an AST parser capable of creating ASTs of compilation units.
	// */
	// public CompilationUnit createRoot(final ICompilationUnit compilationUnit,
	// final IProgressMonitor monitor) {
	// ASTParser astParser = ASTParser.newParser(ASTProvider.SHARED_AST_LEVEL);
	// astParser.setSource(compilationUnit);
	// astParser.setResolveBindings(true);
	// astParser.setStatementsRecovery(ASTProvider.SHARED_AST_STATEMENT_RECOVERY);
	// astParser.setBindingsRecovery(ASTProvider.SHARED_BINDING_RECOVERY);
	// return (CompilationUnit) astParser.createAST(monitor);
	// }
}
