/**
 *
 */
package edu.ncsu.csc.emh.fixbugs.core.dom;

import java.util.List;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.Assignment;
import org.eclipse.jdt.core.dom.BodyDeclaration;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.ExpressionStatement;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.InfixExpression;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.Name;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;
import org.eclipse.jdt.internal.corext.dom.ASTNodes;

/**
 * @see org.eclipse.jdt.internal.corext.dom.ASTNodes
 * @author Yoonki
 */
public final class ASTNodeUtil {

	/**
	 * Checks if the given statement contains '=' (INFIX_EXPRESSION).
	 * 
	 * <pre>
	 * s = s + &quot;abc&quot;;
	 * </pre>
	 * 
	 * @param statement
	 *            the statement.
	 * @return
	 * @return true if the statement contains infix expression; otherwise false.
	 */
	public static boolean containsInfixPlusOperator(final Statement statement) {
		if (!(statement instanceof ExpressionStatement)) {
			return false;
		}
		Expression expr = ((ExpressionStatement) statement).getExpression();
		if (!(expr instanceof Assignment)) {
			return false;
		}
		Assignment assignment = (Assignment) expr;
		Expression lhs = assignment.getLeftHandSide();
		Expression rhs = assignment.getRightHandSide();
		if (lhs.getNodeType() == ASTNode.SIMPLE_NAME && rhs.getNodeType() == ASTNode.INFIX_EXPRESSION) {
			SimpleName name1 = (SimpleName) lhs;
			InfixExpression infixExpr = (InfixExpression) rhs;
			if (infixExpr.getLeftOperand().getNodeType() == ASTNode.SIMPLE_NAME) {
				SimpleName name2 = (SimpleName) infixExpr.getLeftOperand();
				if (isSameNameIdentifier(name1, name2) && infixExpr.getOperator() == InfixExpression.Operator.PLUS) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Checks if the given Statement contains '+=' (PLUS_ASSIGN) operator.
	 * <p>
	 * For example,
	 * 
	 * <pre>
	 * s += &quot;abc&quot;;
	 * </pre>
	 * 
	 * @param stmt
	 *            the statement
	 * @return true if the statement contains plus assignment; otherwise false.
	 */
	public static boolean containsPlusAssignmentOperator(final Statement stmt) {
		if (!(stmt instanceof ExpressionStatement)) {
			return false;
		}
		Expression expr = ((ExpressionStatement) stmt).getExpression();
		if (expr instanceof Assignment) {
			Assignment assignment = (Assignment) expr;
			if (assignment.getOperator() == Assignment.Operator.PLUS_ASSIGN) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Returns the first occurrence of the given variable identifier in
	 * fragments of a variable declaration statement.
	 * 
	 * @param vds
	 * @param identifier
	 * @return
	 */
	public static VariableDeclarationFragment findFirstVariableDeclarationFragment(final VariableDeclarationStatement vds,
			final String identifier) {
		List fragments = vds.fragments();
		for (Object fragment : fragments) {
			VariableDeclarationFragment vdf = (VariableDeclarationFragment) fragment;
			if (identifier.equals(vdf.getName().getIdentifier())) {
				return vdf;
			}
		}
		return null;
	}

	public static Expression findParentExpression(ASTNode node) {
		while ((node != null) && (!(node instanceof Expression))) {
			node = node.getParent();
			if (node instanceof BodyDeclaration) {
				return null;
			}
		}
		return (Expression) node;
	}

	/**
	 * Finds parent loop Statement.
	 * 
	 * @param node
	 *            the seed node
	 * @return parent loop statement.
	 */
	public static Statement findParentLoopStatement(ASTNode node) {
		while ((node != null) && (!isLoopStatement(node))) {
			node = node.getParent();
			if (node instanceof BodyDeclaration) {
				return null;
			}
		}
		return (Statement) node;
	}

	// from ASTResolving
	public static Statement findParentStatement(ASTNode node) {
		while ((node != null) && (!(node instanceof Statement))) {
			node = node.getParent();
			if (node instanceof BodyDeclaration) {
				return null;
			}
		}
		return (Statement) node;
	}

	public static Statement findParentStatementByStartingPosition() {
		return null;
	}

	/**
	 * Finds the right hand operand from the given statement.
	 * 
	 * @param statement
	 *            the statement.
	 * @return the expression found.
	 */
	public static Expression findRighHandOperand(final Statement statement) {
		if (!(statement instanceof ExpressionStatement)) {
			return null;
		}
		Expression expression = ((ExpressionStatement) statement).getExpression();
		if (!(expression instanceof Assignment)) {
			return null;
		}
		Assignment assignment = (Assignment) expression;
		Expression lhs = assignment.getLeftHandSide();
		Expression rhs = assignment.getRightHandSide();
		if (lhs.getNodeType() == ASTNode.SIMPLE_NAME && rhs.getNodeType() == ASTNode.INFIX_EXPRESSION) {
			SimpleName name1 = (SimpleName) lhs;
			InfixExpression infixExpr = (InfixExpression) rhs;
			if (infixExpr.getLeftOperand().getNodeType() == ASTNode.SIMPLE_NAME) {
				SimpleName name2 = (SimpleName) infixExpr.getLeftOperand();
				if (isSameNameIdentifier(name1, name2) && infixExpr.getOperator() == InfixExpression.Operator.PLUS) {
					return infixExpr.getRightOperand();
				}
			}
		}
		return null;
	}

	/**
	 * Find the right hand side of the given statement.
	 * 
	 * @param stmt
	 *            the statement.
	 * @return the expression which is found.
	 */
	public static Expression findRightHandSide(final Statement stmt) {
		if (!(stmt instanceof ExpressionStatement)) {
			return null;
		}
		Expression expr = ((ExpressionStatement) stmt).getExpression();
		if (expr instanceof Assignment) {
			Assignment assignment = (Assignment) expr;
			return assignment.getRightHandSide();
		}
		return null;
	}

	/**
	 * Checks
	 * <ul>
	 * <li>whether the given infix expression is the argument of a
	 * StringBuilder#append() or StringBuffer#append() invocation, and</li>
	 * <li>the append method is called on a simple variable, and</li>
	 * <li>the invocation occurs in a statement (not as nested expression)</li>
	 * </ul>
	 * 
	 * @param infixExpression
	 *            the infix expression
	 * @return the name of the variable we were appending to, or
	 *         <code>null</code> if not matching
	 * @see org.eclipse.jdt.internal.ui.text.correction.QuickAssistProcessor
	 */
	public static SimpleName getEnclosingAppendBuffer(final InfixExpression infixExpression) {
		if (infixExpression.getLocationInParent() == MethodInvocation.ARGUMENTS_PROPERTY) {
			MethodInvocation methodInvocation = (MethodInvocation) infixExpression.getParent();

			// ..not in an expression.. (e.g. not sb.append("high" +
			// 5).append(6);)
			if (methodInvocation.getParent() instanceof Statement) {

				// ..of a function called append:
				if ("append".equals(methodInvocation.getName().getIdentifier())) { //$NON-NLS-1$
					Expression expression = methodInvocation.getExpression();

					// ..and the append is being called on a Simple object:
					if (expression instanceof SimpleName) {
						IBinding binding = ((SimpleName) expression).resolveBinding();
						if (binding instanceof IVariableBinding) {
							String typeName = ((IVariableBinding) binding).getType().getQualifiedName();

							// And the object's type is a StringBuilder or
							// StringBuffer:
							if ("java.lang.StringBuilder".equals(typeName) || "java.lang.StringBuffer".equals(typeName)) { //$NON-NLS-1$ //$NON-NLS-2$
								return (SimpleName) expression;
							}
						}
					}
				}
			}
		}
		return null;
	}

	/**
	 * Checks if the given node is a loop statement.
	 * <p>
	 * Statement:
	 * 
	 * <pre>
	 * Block 
	 * IfStatement 
	 * ForStatement 
	 * EnhancedForStatement
	 * WhileStatement 
	 * DoStatement 
	 * TryStatement 
	 * SwitchStatement
	 * SynchronizedStatement 
	 * ReturnStatement 
	 * ThrowStatement 
	 * BreakStatement
	 * ContinueStatement 
	 * EmptyStatement 
	 * ExpressionStatement 
	 * LabeledStatement
	 * AssertStatement 
	 * VariableDeclarationStatement 
	 * TypeDeclarationStatement
	 * ConstructorInvocation 
	 * SuperConstructorInvocation
	 * </pre>
	 * 
	 * @param node
	 *            the AST node.
	 * @return true if the node is a loop statement; false otherwise.
	 */
	public static boolean isLoopStatement(final ASTNode node) {
		int type = node.getNodeType();
		if (type == ASTNode.FOR_STATEMENT || type == ASTNode.ENHANCED_FOR_STATEMENT || type == ASTNode.WHILE_STATEMENT
				|| type == ASTNode.DO_STATEMENT) {
			return true;
		}
		return false;
	}

	public static boolean isControlStatement(ASTNode node) {
		int type = node.getNodeType();
		return type == ASTNode.IF_STATEMENT || isLoopStatement(node);
	}

	public static boolean isControlStatementWithBlock(ASTNode node) {
		switch (node.getNodeType()) {
		case ASTNode.IF_STATEMENT:
		case ASTNode.WHILE_STATEMENT:
		case ASTNode.FOR_STATEMENT:
		case ASTNode.ENHANCED_FOR_STATEMENT:
		case ASTNode.DO_STATEMENT:
			return true;
		default:
			return false;
		}
	}

	private static boolean isSameNameIdentifier(final Name name1, final Name name2) {
		if (ASTNodes.getSimpleNameIdentifier(name1).equals(ASTNodes.getSimpleNameIdentifier(name2))) {
			return true;
		}
		return false;
	}

	public static int getIndex(int offset, List<Statement> statements) {
		for (int i = 0; i < statements.size(); i++) {
			Statement s = statements.get(i);
			if (offset <= s.getStartPosition()) {
				return i;
			}
			if (offset < s.getStartPosition() + s.getLength()) {
				return -1;
			}
		}
		return statements.size();
	}

	private ASTNodeUtil() {
		// no instance
	}
}
