//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.core.resolution;

import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.runtime.Assert;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.BodyDeclaration;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.NodeFinder;
import org.eclipse.jdt.core.dom.rewrite.ASTRewrite;
import org.eclipse.jdt.core.dom.rewrite.ITrackedNodePosition;
import org.eclipse.jdt.internal.ui.text.correction.ASTResolving;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Document;
import org.eclipse.jface.text.Position;
import org.eclipse.text.edits.TextEdit;

import edu.ncsu.csc.emh.fixbugs.core.annotation.FixBugsAnnotation;

/**
 * 
 * @see org.eclipse.jdt.internal.ui.text.correction.Assist-Context (remove '-')
 * @author Yoonki Song
 */
@SuppressWarnings("restriction")
public abstract class FixBugsAssistant implements IAnnotationSupport {

	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(FixBugsAssistant.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	/**
	 * Constructor for FixBugsAssistant.
	 * 
	 * @param context
	 *            the invocation context.
	 */
	public FixBugsAssistant(final FixBugsContext context) {
		fContext = context;
	}

	/** */
	private final FixBugsContext fContext;

	/** */
	private Position fAdjustedSelection;

	/** */
	private Document fDocument = null;

	/** */
	private String fDocumentAsString = null;

	/** */
	private BodyDeclaration fCoveringBodyDecl = null;

	/** */
	private Position fCoveringBodyDeclPosition;

	/** */
	private String fCoveringBodyDeclAsString = null;

	/** */
	private String fCoveringCUAsString = null;

	/**
	 * Saves the very initial status.
	 * <p>
	 * <b>NOTE</b> It is important to keep the original information such as the
	 * text of the covering body declaration because the body declaration itself
	 * may be changed during the IQF operations.
	 * 
	 * @throws BadLocationException
	 *             is thrown if one of the edits in the tree can't be executed.
	 *             The state of the document is undefined if this exception is
	 *             thrown.
	 * @throws JavaModelException
	 *             if an error occurs in the Java model.
	 */
	public void init() throws BadLocationException, JavaModelException {
		getDocument();
		getDocumentAsString();
		getCoveringCUAsString();
		getCoveringBodyDecl();
		getCoveringBodyDeclPosition();
		getCoveringBodyDeclAsString();
	}

	/**
	 * Gets the invocation context for FixBugs.
	 * 
	 * @return the invocation context.
	 */
	public final FixBugsContext getContext() {
		return fContext;
	}

	/**
	 * Returns the document.
	 * 
	 * @return the document.
	 * @throws JavaModelException
	 *             if an error occurs in the Java model.
	 */
	public final Document getDocument() throws JavaModelException {
		if (fDocument == null) {
			fDocument = new Document(fContext.getCompilationUnit().getSource());
		}
		return fDocument;
	}

	/**
	 * Returns the cached text in document as string.
	 * 
	 * @return the cached text in document as string.
	 * @throws JavaModelException
	 *             if an error occurs in the Java model.
	 */
	public final String getDocumentAsString() throws JavaModelException {
		if (fDocumentAsString == null) {
			Document document = getDocument();
			fDocumentAsString = document.get();
		}
		return fDocumentAsString;
	}

	/**
	 * Returns the adjusted selection.
	 * 
	 * @return the adjusted selection.
	 */
	public final Position getAdjustedSelection() {
		if (fAdjustedSelection == null) {
			fAdjustedSelection = getInitialSelection();
		}

		return fAdjustedSelection;
	}

	/**
	 * This method should be called once when opening the interactive quick fix
	 * popup.
	 * 
	 * @param adjustedSelection
	 *            the adjusted selection.
	 */
	public final void setAdjustedSelection(final Position adjustedSelection) {
		fAdjustedSelection = adjustedSelection;
	}

	/**
	 * This method should be called once when opening the interactive quick fix
	 * popup.
	 * 
	 * @param offset
	 *            the offset of selected position.
	 * @param length
	 *            the length of selected position.
	 */
	public final void setAdjustedSelection(final int offset, final int length) {
		fAdjustedSelection = new Position(offset, length);
	}

	/**
	 * Returns the selected position.
	 * 
	 * @return the selected position.
	 */
	public final Position getInitialSelection() {
		return new Position(fContext.getSelectionOffset(), fContext.getSelectionLength());
	}

	/**
	 * Returns covering body declaration.
	 * 
	 * YS Can be more precise AST Node rather than body declaration (such as
	 * method declaration). It may depend on the solution of a bug pattern.
	 * 
	 * @return the enclosing body declaration.
	 */
	public final BodyDeclaration getCoveringBodyDecl() {
		if (fCoveringBodyDecl == null) {
			fCoveringBodyDecl = ASTResolving.findParentBodyDeclaration(fContext.getCoveringNode());
		}

		return fCoveringBodyDecl;
	}

	/**
	 * Returns covering CompilationUnit as String for cache purpose.
	 * 
	 * @return the string representation of CompilationUnit.
	 * @throws JavaModelException
	 *             if an error occurs in the Java model
	 * @throws BadLocationException
	 *             is thrown if one of the edits in the tree can't be executed.
	 *             The state of the document is undefined if this exception is
	 *             thrown.
	 */
	public final String getCoveringCUAsString() throws JavaModelException, BadLocationException {
		if (fCoveringCUAsString == null) {
			Document document = getDocument();
			CompilationUnit root = fContext.getASTRoot();
			fCoveringCUAsString = document.get(root.getStartPosition(), root.getLength());
		}

		return fCoveringCUAsString;
	}

	/**
	 * Gets the covering BodyDeclaration as string for caching purpose.
	 * 
	 * @return the covering BodyDeclaration as string for caching purpose.
	 * @throws JavaModelException
	 *             if an error occurs in the Java model
	 * @throws BadLocationException
	 *             is thrown if one of the edits in the tree can't be executed.
	 *             The state of the document is undefined if this exception is
	 *             thrown.
	 */
	public final String getCoveringBodyDeclAsString() throws JavaModelException, BadLocationException {
		if (fCoveringBodyDeclAsString == null) {
			Document document = getDocument();
			BodyDeclaration bd = getCoveringBodyDecl();
			fCoveringBodyDeclAsString = document.get(bd.getStartPosition(), bd.getLength());
		}

		return fCoveringBodyDeclAsString;
	}

	/**
	 * Gets the covering position of BodyDeclaration.
	 * 
	 * @return the covering position of BodyDeclaration.
	 */
	public final Position getCoveringBodyDeclPosition() {
		if (fCoveringBodyDeclPosition == null) {
			BodyDeclaration bd = getCoveringBodyDecl();
			fCoveringBodyDeclPosition = new Position(bd.getStartPosition(), bd.getLength());
		}
		return fCoveringBodyDeclPosition;
	}

	/**
	 * Gets the covering BodyDeclaration.
	 * 
	 * @param selection
	 *            the selection.
	 * @return the covering BodyDeclaration.
	 */
	public final BodyDeclaration getCoveringBodyDecl(final Position selection) {
		return ASTResolving.findParentBodyDeclaration(fContext.getCoveringNode());
	}

	/**
	 * Finds the AST Node in document.
	 * 
	 * @param offset
	 *            the offset.
	 * @return the AST Node which is found.
	 */
	public final ASTNode findASTNode(final int offset, final int length) {
		NodeFinder nf = new NodeFinder(fContext.getASTRoot(), offset, length);
		return nf.getCoveringNode();
	}

	/**
	 * Finds the AST Node in document.
	 * 
	 * @param position
	 *            the position.
	 * @return the AST Node which is found.
	 */
	public final ASTNode findASTNode(final Position position) {
		return findASTNode(position.getOffset(), position.getLength());
	}

	/** */
	private ASTNode[] fAnnotationNodes = null;

	/** */
	private List<FixBugsAnnotation> fAnnotations = null;

	@Override
	public final ASTNode[] getAnnotationNodes() {
		return fAnnotationNodes;
	}

	@Override
	public final void setAnnotationNodes(final ASTNode[] annotationNodes) {
		fAnnotationNodes = annotationNodes;
	}

	/**
	 * Returns the annotation of FixBugs.
	 * 
	 * @return the annotations.
	 */
	@Override
	public final List<FixBugsAnnotation> getAnnotations() {
		return fAnnotations;
	}

	/**
	 * Finds an overlapped annotation.
	 * 
	 * @param offset
	 *            the document offset
	 * @return the annotation which is found.
	 */
	@Override
	public final FixBugsAnnotation findAnnotation(final int offset) {
		List<FixBugsAnnotation> annotations = getAnnotations();
		for (FixBugsAnnotation annotation : annotations) {
			// fLogger.fine("suggestible anno: " + annotation);
			if (!annotation.getPosition().isDeleted() /* important */
					&& annotation.includes(offset)) {
				return annotation;
			}
		}
		return null;
	}

	/**
	 * Sets the annotations of FixBugs.
	 * 
	 * @param annotations
	 *            new annotations.
	 */
	@Override
	public final void setAnnotations(final List<FixBugsAnnotation> annotations) {
		fAnnotations = annotations;
	}

	/**
	 * Gets the position of the given ASTNode.
	 * 
	 * @param node
	 *            the ASTNode
	 * @return the Position.
	 */
	protected static Position getNodePosition(final ASTNode node) {
		Assert.isNotNull(node);
		return new Position(node.getStartPosition(), node.getLength());
	}

	/**
	 * Gets the MethodDeclaration under selection.
	 * 
	 * @param offset
	 *            the offset
	 * @param length
	 *            the length
	 * @return the MethodDeclaration
	 */
	public final MethodDeclaration getMethodUnderPosition(final int offset, final int length) {
		ASTNode node = findASTNode(offset, length);
		while ((node != null) && !(node instanceof MethodDeclaration)) {
			node = node.getParent();
			if (node instanceof MethodDeclaration) {
				return (MethodDeclaration) node;
			}
		}
		return null;
	}

	protected String rewriteMethodDecl(final ASTRewrite rewrite) throws BadLocationException, JavaModelException {
		// 1. Apply the change to the original unit and get new text from it
		// creation of a Document
		Document document = new Document(getDocumentAsString());

		BodyDeclaration bodyDeclaration = getCoveringBodyDecl();
		ITrackedNodePosition trackedBodyDeclaration = rewrite.track(bodyDeclaration);

		// computation of the text edits
		ICompilationUnit originalUnit = getContext().getCompilationUnit();
		TextEdit edit = rewrite.rewriteAST(document, originalUnit.getJavaProject().getOptions(true));
		edit.apply(document, TextEdit.UPDATE_REGIONS);

		// get new text from the updated compilation unit
		return document.get(trackedBodyDeclaration.getStartPosition(), trackedBodyDeclaration.getLength());
	}

	// scope: whole compilation unit
	protected String rewriteCU(final ASTRewrite rewrite) throws BadLocationException, JavaModelException {
		// 1. Apply the change to the original unit and get new text from it
		// creation of a Document
		Document document = new Document(getDocumentAsString());

		CompilationUnit root = getContext().getASTRoot();
		ITrackedNodePosition trackedRoot = rewrite.track(root);

		// computation of the text edits
		ICompilationUnit originalUnit = getContext().getCompilationUnit();
		TextEdit edit = rewrite.rewriteAST(document, originalUnit.getJavaProject().getOptions(true));
		edit.apply(document, TextEdit.UPDATE_REGIONS);

		// get new text from the updated compilation unit
		return document.get(trackedRoot.getStartPosition(), trackedRoot.getLength());
	}

	/**
	 * Does node1 include node2?
	 * 
	 * @param node1
	 * @param node2
	 * @return
	 */
	protected boolean includes(final ASTNode node1, final ASTNode node2) {
		if (node1.getStartPosition() < node2.getStartPosition()
				&& (node1.getStartPosition() + node1.getLength() > node2.getStartPosition() + node2.getLength())) {
			return true;
		}
		return false;
	}

	/**
	 * 
	 * @param node
	 * @return
	 */
	protected final boolean isSelectionOverlapped(final ASTNode node) {
		if (node.getStartPosition() >= getContext().getSelectionOffset()) {
			return true;
		}
		return false;
	}

	protected ASTRewrite fRewrite = null;

	public ASTRewrite getASTRewrite() {
		return fRewrite;
	}
}
