//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.core.resolution;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.NodeFinder;
import org.eclipse.jdt.internal.ui.javaeditor.JavaEditor;
import org.eclipse.jdt.ui.SharedASTProvider;
import org.eclipse.jdt.ui.text.java.IInvocationContext;
import org.eclipse.jface.text.ITextSelection;
import org.eclipse.jface.text.source.ISourceViewer;
import org.eclipse.jface.text.source.TextInvocationContext;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionProvider;
import org.eclipse.ui.IEditorPart;

/**
 * 
 * @see org.eclipse.jdt.internal.ui.text.correction.Assist-Context (remove '-')
 * @author Yoonki Song
 */
public class FixBugsContext extends TextInvocationContext implements IInvocationContext {
	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(FixBugsContext.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	/** */
	private final ICompilationUnit fCompilationUnit;

	/** */
	private final IEditorPart fEditorPart;

	/** */
	private CompilationUnit fASTRoot;

	/** */
	private final SharedASTProvider.WAIT_FLAG fWaitFlag;

	/**
	 * The cached node finder, can be null.
	 * 
	 * @since 3.6
	 */
	private NodeFinder fNodeFinder;

	/*
	 * @since 3.5
	 */
	private FixBugsContext(final ICompilationUnit cu, final ISourceViewer sourceViewer, final IEditorPart editorPart,
			final int offset, final int length, final SharedASTProvider.WAIT_FLAG waitFlag) {
		super(sourceViewer, offset, length);
		Assert.isLegal(cu != null);
		Assert.isLegal(waitFlag != null);
		fCompilationUnit = cu;
		fEditorPart = editorPart;
		fWaitFlag = waitFlag;
	}

	/*
	 * @since 3.5
	 */
	public FixBugsContext(final ICompilationUnit cu, final ISourceViewer sourceViewer, final int offset, final int length,
			final SharedASTProvider.WAIT_FLAG waitFlag) {
		this(cu, sourceViewer, null, offset, length, waitFlag);
	}

	/*
	 * @since 3.5
	 */
	public FixBugsContext(final ICompilationUnit cu, final ISourceViewer sourceViewer, final IEditorPart editor,
			final int offset, final int length) {
		this(cu, sourceViewer, editor, offset, length, SharedASTProvider.WAIT_YES);
	}

	/*
	 * Constructor for CorrectionContext.
	 * 
	 * @since 3.4
	 */
	public FixBugsContext(final ICompilationUnit cu, final ISourceViewer sourceViewer, final int offset, final int length) {
		this(cu, sourceViewer, null, offset, length);
	}

	/**
	 * Constructor for FixInvocationContext.
	 * 
	 * @param cu
	 *            the compilation unit.
	 * @param offset
	 *            the offset.
	 * @param length
	 *            the length.
	 */
	public FixBugsContext(final ICompilationUnit cu, final int offset, final int length) {
		this(cu, null, offset, length);
	}

	/**
	 * Returns the compilation unit.
	 * 
	 * @return an <code>ICompilationUnit</code>
	 */
	@Override
	public final ICompilationUnit getCompilationUnit() {
		return fCompilationUnit;
	}

	/**
	 * Returns the editor or <code>null</code> if none.
	 * 
	 * @return an <code>IEditorPart</code> or <code>null</code> if none
	 * @since 3.5
	 */
	public final IEditorPart getEditorPart() {
		return fEditorPart;
	}

	/**
	 * Returns the length.
	 * 
	 * @return the length.
	 */
	@Override
	public final int getSelectionLength() {
		return Math.max(getLength(), 0);
	}

	/**
	 * Returns the offset.
	 * 
	 * @return the offset.
	 */
	@Override
	public final int getSelectionOffset() {
		return getOffset();
	}

	@Override
	public final CompilationUnit getASTRoot() {
		if (fASTRoot == null) {
			fASTRoot = SharedASTProvider.getAST(fCompilationUnit, fWaitFlag, null);
			if (fASTRoot == null) {
				// see bug 63554
				fASTRoot = createQuickFixAST(fCompilationUnit, null);
			}
		}
		return fASTRoot;
	}

	/** */
	private CompilationUnit createQuickFixAST(final ICompilationUnit compilationUnit, final IProgressMonitor monitor) {
		ASTParser astParser = ASTParser.newParser(AST.JLS4);
		astParser.setSource(compilationUnit);
		astParser.setResolveBindings(true);
		astParser.setStatementsRecovery(true);
		astParser.setBindingsRecovery(true);
		return (CompilationUnit) astParser.createAST(monitor);
	}

	/**
	 * @param root
	 *            The ASTRoot to set.
	 */
	public final void setASTRoot(final CompilationUnit root) {
		fASTRoot = root;
	}

	@Override
	public final ASTNode getCoveringNode() {
		if (fNodeFinder == null) {
			fNodeFinder = new NodeFinder(getASTRoot(), getOffset(), getLength());
		}
		return fNodeFinder.getCoveringNode();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jdt.ui.text.java.IInvocationContext#getCoveredNode()
	 */
	@Override
	public final ASTNode getCoveredNode() {
		if (fNodeFinder == null) {
			fNodeFinder = new NodeFinder(getASTRoot(), getOffset(), getLength());
		}
		return fNodeFinder.getCoveredNode();
	}

	public ITextSelection getTextSelection() {
		fLogger.info("#getTextSelection");
		JavaEditor activeEditor = (JavaEditor) getEditorPart();
		Assert.isNotNull(activeEditor);
		ISelectionProvider selectionProvider = activeEditor.getSelectionProvider();
		if (selectionProvider == null) {
			return null;
		}

		ISelection selection = selectionProvider.getSelection();
		if (!(selection instanceof ITextSelection)) {
			return null;
		}

		return (ITextSelection) selection;
	}

}
