//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.core.resolution.correctness.missingArgument;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.eclipse.core.runtime.Assert;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.ExpressionStatement;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.QualifiedName;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.StringLiteral;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.rewrite.ASTRewrite;
import org.eclipse.jdt.core.dom.rewrite.ITrackedNodePosition;
import org.eclipse.jdt.internal.corext.dom.ScopeAnalyzer;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Position;

import edu.ncsu.csc.emh.fixbugs.core.annotation.FixBugsAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.annotation.VariableAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsAssistant;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsContext;
import edu.ncsu.csc.emh.fixbugs.core.util.Pair;
import edu.ncsu.csc.emh.fixbugs.core.util.SignatureUtil;
import edu.ncsu.csc.emh.fixbugs.ui.Messages;

/**
 * 
 * @author Yoonki Song
 */
@SuppressWarnings("restriction")
public final class MissingArgumentAssistant extends FixBugsAssistant {

	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(MissingArgumentAssistant.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	/** missing argument index in the string literal. */
	private static int MISSING_ARG_INDEX = -1;

	/** from JDK 7 code. */
	private static final String formatSpecifier = "%(\\d+\\$)?([-#+ 0,(\\<]*)?(\\d+)?(\\.\\d+)?([tT])?([a-zA-Z%])";

	/** */
	private static Pattern fsPattern = Pattern.compile(formatSpecifier);

	private static StringLiteral findBestGuessedStringLiteralNode(final ASTNode node) {
		if (node instanceof ExpressionStatement) {
			// e.g. System.out.printf("%s");
			Expression expr = ((ExpressionStatement) node).getExpression();
			if (expr instanceof MethodInvocation) {
				List arguments = ((MethodInvocation) expr).arguments();
				for (Object argument : arguments) {
					ASTNode argNode = (ASTNode) argument;
					if (argNode.getNodeType() == ASTNode.STRING_LITERAL) {
						return (StringLiteral) argNode;
					}
				}
			}
		}
		return null;
	}

	private static SimpleName findSimpleName(ASTNode node) {
		SimpleName sn = null;
		if (node instanceof VariableDeclarationFragment) {
			sn = ((VariableDeclarationFragment) node).getName();
		} else if (node instanceof SingleVariableDeclaration) {
			sn = ((SingleVariableDeclaration) node).getName();
		} // YS anything else??
		return sn;
	}

	private static List<IVariableBinding> findSuggestibleVariableBindings(final CompilationUnit root, final Position position,
			final String[] types) {
		ScopeAnalyzer analyzer = new ScopeAnalyzer(root);
		IBinding[] bindings = analyzer.getDeclarationsInScope(position.getOffset(), ScopeAnalyzer.VARIABLES);
		List<IVariableBinding> suggestibleBindings = new ArrayList<IVariableBinding>();
		for (int i = 0; i < bindings.length; i++) {
			IVariableBinding curr = (IVariableBinding) bindings[i];
			ITypeBinding type = curr.getType();
			if (type != null && matches(type.getQualifiedName(), types)) {
				suggestibleBindings.add(curr);
			}
		}

		return suggestibleBindings;
	}

	private static List<SimpleName> findSuggestibleVariables(final CompilationUnit root,
			final List<IVariableBinding> variableBindings) {
		// sorted by name
		List<SimpleName> suggestibles = new ArrayList<SimpleName>();
		for (IVariableBinding vb : variableBindings) {
			ASTNode node = root.findDeclaringNode(vb);
			SimpleName sn = findSimpleName(node);
			if (sn != null) {
				fLogger.fine("Suggestible Variable: " + sn);
				suggestibles.add(sn);
			}
		}
		return suggestibles;
	}

	/**
	 * Finds default variable suggestion among suggestible variables.
	 * 
	 * @param root
	 * @param baseNode
	 *            the base AST node. it is used for comparing distances between
	 *            the node and suggestible variables.
	 * @param variableBindings
	 *            the variable bindings of variables.
	 * @return
	 */
	private static SimpleName findVariableSuggestion(final CompilationUnit root, final ASTNode baseNode,
			final List<IVariableBinding> variableBindings) {
		fLogger.info("#findVariableSuggestion");
		SimpleName selected = null;
		// look through previously defined local/parameter variables first
		List<ASTNode> suggestibles = new ArrayList<ASTNode>();
		for (IVariableBinding vb : variableBindings) {
			ASTNode node = root.findDeclaringNode(vb);
			if (node.getStartPosition() < baseNode.getStartPosition()) {
				if (!vb.isField()) { // parameter or local
					suggestibles.add(node);
				}
			}
		}
		// look through previously defined field variables
		if (suggestibles.isEmpty()) { // field
			for (IVariableBinding vb : variableBindings) {
				ASTNode node = root.findDeclaringNode(vb);
				if (node.getStartPosition() < baseNode.getStartPosition()) {
					if (vb.isField()) { // field
						suggestibles.add(node);
					}
				}
			}
		}
		// in this case, collect all.
		if (suggestibles.isEmpty()) {
			for (IVariableBinding vb : variableBindings) {
				ASTNode node = root.findDeclaringNode(vb);
				if (node.getStartPosition() > baseNode.getStartPosition()) {
					suggestibles.add(node);
				}
			}
		}

		// finally, if there are suggestible variables!
		if (!suggestibles.isEmpty()) {
			Collections.sort(suggestibles, new Comparator<ASTNode>() {
				@Override
				public int compare(final ASTNode n1, final ASTNode n2) {
					int distance1 = baseNode.getStartPosition() - n1.getStartPosition();
					if (distance1 < 0) {
						distance1 = -distance1;
					}
					int distance2 = baseNode.getStartPosition() - n2.getStartPosition();
					if (distance2 < 0) {
						distance2 = -distance2;
					}
					return distance1 - distance2;
				}
			});
			selected = findSimpleName(suggestibles.get(0));
		}
		return selected;
	}

	/** */
	private static String[] getConversionCharacters(final String formatString) {
		// System.out.println("FormatString: " + formatString);
		Matcher matcher = fsPattern.matcher(formatString);
		List<String> conversionList = new ArrayList<String>();
		while (matcher.find()) {
			conversionList.add(matcher.group());
		}
		return conversionList.toArray(new String[conversionList.size()]);
	}

	private static boolean matches(final String qualifiedName, final String[] types) {
		for (String type : types) {
			if (qualifiedName.contains(type)) {
				return true;
			}
		}
		return false;
	}

	/** */
	private StringLiteral fStringLiteral;

	/** suggestible variables. */
	private List<SimpleName> fSuggestibleVariables = null;

	/** Selected Variable among suggestible variables. */
	private SimpleName fSelectedVariable = null;

	/** */
	private ASTNode fLastExistingArgument = null;

	/** */
	private Position fMissingArgumentPosition;

	/** */
	private Position fPresentedVariablePosition;

	/** */
	private FixBugsAnnotation fPresentedVariableAnnotation = null;

	/** */
	private String fPresentedVariableName = null;

	/**
	 * Constructor for MissingArgumentAssistant.
	 * 
	 * @param context
	 *            the invocation context.
	 */
	public MissingArgumentAssistant(final FixBugsContext context) {
		super(context);
	}

	private Position calcuateMissingArugmentPosition(StringLiteral stringLiteral, String missingArgument) {
		String escapedValue = stringLiteral.getEscapedValue();
		int index = escapedValue.indexOf(missingArgument);
		return new Position(stringLiteral.getStartPosition() + index, missingArgument.length());
	}

	/**
	 * Applies the ...
	 * 
	 * @param nameToPresent
	 *            the variable name to present.
	 * @param trackedStrLitPair
	 *            tracked annotation map.
	 * @param trackedPresentedVariablePair
	 *            tracked variable annotation map.
	 * @param trackedSuggestibleVariableMap
	 *            tracked suggestible variables.
	 * @return the changed text
	 * 
	 * @throws BadLocationException
	 *             is thrown if one of the edits in the tree can't be executed.
	 *             The state of the document is undefined if this exception is
	 *             thrown.
	 * @throws JavaModelException
	 *             the exception is thrown if the rewrite fails.
	 */
	public String changeAST(String nameToPresent, Pair<ITrackedNodePosition, FixBugsAnnotation> trackedStrLitPair,
			Pair<ITrackedNodePosition, FixBugsAnnotation> trackedPresentedVariablePair,
			final Map<ITrackedNodePosition, FixBugsAnnotation> trackedSuggestibleVariableMap) throws BadLocationException,
			JavaModelException {
		fLogger.info("MissingArgumentAssistant#changeAST");

		// get DOM/AST from a ICompilationUnit
		CompilationUnit root = getContext().getASTRoot();

		// creation of ASTRewrite
		AST ast = root.getAST();
		ASTRewrite rewrite = ASTRewrite.create(ast);
		ICompilationUnit cu = getContext().getCompilationUnit();

		// ----------------------------- Change -----------------------------//

		ASTNode[] annotationNodes = getAnnotationNodes();

		// description of the change
		ASTNode strLitNode = annotationNodes[0];
		fLogger.info("\tstrLitNode: " + strLitNode);
		ASTNode parentNode = strLitNode.getParent();
		if (parentNode instanceof MethodInvocation) {
			// e.g. System.out.printf("%s");
			MethodInvocation curMI = ((MethodInvocation) parentNode);
			MethodInvocation newMI = (MethodInvocation) ASTNode.copySubtree(ast, curMI);
			List arguments = newMI.arguments();
			SimpleName argumentToPresent = ast.newSimpleName(nameToPresent);
			arguments.add(argumentToPresent);
			rewrite.replace(curMI, newMI, null);
			// ex) i in System.out.println("%d", i);
			FixBugsAnnotation strLitAnnotation = new FixBugsAnnotation(FixBugsAnnotation.MISSING_ARGUMENT); // initial
			trackedStrLitPair.setLeft(rewrite.track(strLitNode));
			trackedStrLitPair.setRight(strLitAnnotation);
			// ex) "%d" in System.out.println("%d", i);
			FixBugsAnnotation presentedVariableAnnotation = new FixBugsAnnotation(FixBugsAnnotation.KIND1P); // initial
			presentedVariableAnnotation.setText(Messages.presentedVariableAnnotation);
			presentedVariableAnnotation.setName(argumentToPresent.getIdentifier());
			trackedPresentedVariablePair.setLeft(rewrite.track(argumentToPresent));
			trackedPresentedVariablePair.setRight(presentedVariableAnnotation);
		}

		for (SimpleName varName : fSuggestibleVariables) {
			FixBugsAnnotation variableAnnotation = new VariableAnnotation(FixBugsAnnotation.KIND1P); // initial
			variableAnnotation.setText(Messages.suggestibleVariableAnnotation);
			variableAnnotation.setName(varName.getIdentifier());
			trackedSuggestibleVariableMap.put(rewrite.track(varName), variableAnnotation);
		}

		// when the unit is written on disk (and the linked mode is
		// updated), the popup should be opened!

		fRewrite = rewrite; // (important for quickfix)

		return rewriteMethodDecl(rewrite);
	}

	/**
	 * Finds the first missing argument and the last existing argument. For
	 * example,
	 * 
	 * <pre>
	 * System.out.println(&quot;%s %.3f&quot;, s);
	 * </pre>
	 * 
	 * or
	 * 
	 * <pre>
	 * System.out.println(Locale, &quot;%s %.3f&quot;, s);
	 * </pre>
	 * 
	 * %.3f
	 * 
	 * @return the first missing argument.
	 */
	private String findFirstMissingArugment(StringLiteral strlit, boolean isSuggestionMode) {
		String[] conversions = getConversionCharacters(strlit.getEscapedValue());
		ASTNode parent = strlit.getParent();
		if (parent instanceof MethodInvocation) {
			MethodInvocation mi = (MethodInvocation) parent;
			int totalSize = mi.arguments().size();
			int existingArgSize = totalSize - 1; // (format-string)
			if (mi.arguments().get(0) instanceof QualifiedName) { // LOCALE
				existingArgSize--; // (LOCALE)
			}
			MISSING_ARG_INDEX = existingArgSize;
			fLastExistingArgument = (ASTNode) mi.arguments().get(totalSize - 1);
			fLogger.warning("(" + isSuggestionMode + ") Total Argument Size: " + totalSize + ", Existing Argument Size: "
					+ existingArgSize + ", MISSING_ARGUMENT_INDEX: " + MISSING_ARG_INDEX);
		}
		fLogger.warning(conversions[MISSING_ARG_INDEX] + " in conversions: " + Arrays.toString(conversions));
		return conversions[MISSING_ARG_INDEX];
	}

	/**
	 * Finds the string literal that has missing argument(s).
	 * 
	 * @param nodeType
	 *            the node type to find
	 * @param node
	 *            the seed node to find the StringLiteral
	 * @return
	 */
	private StringLiteral findStringLiteral(final ASTNode node) {
		StringLiteral seedNode = null;
		if (node instanceof Block) {
			List<Statement> statements = ((Block) node).statements();
			for (Statement statement : statements) {
				// the nearest statement to the current selection
				if (isSelectionOverlapped(statement)) {
					seedNode = findBestGuessedStringLiteralNode(statement);
					break;
				}
			}
		} else if (node instanceof ExpressionStatement) {
			seedNode = findBestGuessedStringLiteralNode(node);
		} else if (node instanceof StringLiteral) {
			seedNode = (StringLiteral) node;
		} else {
			fLogger.warning("Not supported!");
		}
		return seedNode;
	}

	public Position getLastExistingArgumentPosition() {
		return getNodePosition(fLastExistingArgument);
	}

	public Position getMissingArgumentPosition() {
		return fMissingArgumentPosition;
	}

	public FixBugsAnnotation getPresentedVariableAnnotation() {
		return fPresentedVariableAnnotation;
	}

	public String getPresentedVariableName() {
		return fPresentedVariableName;
	}

	public SimpleName getSelectedVariable() {
		return fSelectedVariable;
	}

	public Position getPresentedVariablePosition() {
		return fPresentedVariablePosition;
	}

	/**
	 * Returns set of suggestible variables.
	 * 
	 * @return set of suggestible variables.
	 */
	public List<SimpleName> getSuggestibleVariables() {
		return fSuggestibleVariables;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void init() throws BadLocationException, JavaModelException {
		super.init();
		initVariableInfo();
		initAnnotations();
	}

	private void initVariableInfo() {
		fStringLiteral = findStringLiteral(getContext().getCoveringNode());

		// find suggestible variables and 1 variable suggestion among them based
		// on policy
		String missingArgument = findFirstMissingArugment(fStringLiteral, false);

		// calculate the missing argument position
		fMissingArgumentPosition = calcuateMissingArugmentPosition(fStringLiteral, missingArgument);
		// System.out.println("Missing Argument Position: " +
		// fMissingArgumentPosition);

		CompilationUnit root = getContext().getASTRoot();
		String[] types = SignatureUtil.findPossibleTypes(missingArgument);
		fLogger.info("Possible Types for the missing argument: " + Arrays.toString(types));
		List<IVariableBinding> variableBindings = findSuggestibleVariableBindings(root, getInitialSelection(), types);
		fSuggestibleVariables = findSuggestibleVariables(root, variableBindings);
		fSelectedVariable = findVariableSuggestion(root, fStringLiteral, variableBindings);
		fLogger.info("Suggestible Variables: " + fSuggestibleVariables.toString());
		fLogger.info("Selected Variable: " + fSelectedVariable.toString());

		// setAdjustedSelection(getNodePosition(fStringLiteral));
		setAdjustedSelection(fMissingArgumentPosition);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void initAnnotations() {
		initAnnotationNodes();
		initAnnotationTypes();
	}

	private void initAnnotationNodes() {
		Assert.isNotNull(fStringLiteral); // precondition
		List<ASTNode> annotationNodes = new ArrayList<ASTNode>();
		annotationNodes.add(fStringLiteral);
		annotationNodes.addAll(fSuggestibleVariables);
		setAnnotationNodes(annotationNodes.toArray(new ASTNode[annotationNodes.size()]));
	}

	private void initAnnotationTypes() {
		List<FixBugsAnnotation> annotations = new ArrayList<FixBugsAnnotation>();
		FixBugsAnnotation missingArgument = new FixBugsAnnotation(FixBugsAnnotation.MISSING_ARGUMENT,
				Messages.missingArgumentAnnotation, fMissingArgumentPosition);
		annotations.add(missingArgument);

		for (SimpleName variable : fSuggestibleVariables) {
			Position p = new Position(variable.getStartPosition(), variable.getLength());
			// highlight body
			FixBugsAnnotation variableAnnotation = new VariableAnnotation(FixBugsAnnotation.KIND1P, p); // initial
			variableAnnotation.setText(Messages.suggestibleVariableAnnotation);
			variableAnnotation.setName(variable.getIdentifier());
			annotations.add(variableAnnotation);

			// border
			FixBugsAnnotation draggableAnnotation = new FixBugsAnnotation(FixBugsAnnotation.DRAGGABLE, p); // initial
			annotations.add(draggableAnnotation);
		}
		setAnnotations(annotations);
	}

	public boolean isPresentedVariableName(String name) {
		Assert.isNotNull(name);
		return name.equals(fPresentedVariableName);
	}

	public void setPresentedVariableAnnotation(FixBugsAnnotation annotation) {
		fPresentedVariableAnnotation = annotation;
	}

	public void setPresentedVariableName(String variableName) {
		fPresentedVariableName = variableName;
	}

	/**
	 * Updates ...
	 * 
	 * @param missingArgument
	 * @param presentedVariable
	 * @param suggestibleVariables
	 */
	public void update(final FixBugsAnnotation missingArgument, final FixBugsAnnotation presentedVariable,
			final List<FixBugsAnnotation> suggestibleVariables) {
		long startTime = System.currentTimeMillis();

		fPresentedVariableAnnotation = presentedVariable;
		fPresentedVariableName = presentedVariable.getName();
		fPresentedVariablePosition = presentedVariable.getPosition();

		// update annotations
		final List<FixBugsAnnotation> annotations = new ArrayList<FixBugsAnnotation>();
		annotations.add(missingArgument);

		// presented variable (highlight body)
		annotations.add(presentedVariable);
		// presented variable (border)
		FixBugsAnnotation presentedVariableBorder = new FixBugsAnnotation(FixBugsAnnotation.SELECTED,
				presentedVariable.getPosition());
		annotations.add(presentedVariableBorder);

		// suggestible variables (highlight body)
		annotations.addAll(suggestibleVariables);

		// suggestible variables (border)
		for (FixBugsAnnotation suggestible : suggestibleVariables) {
			// System.out.println("suggestible var name: " +
			// suggestible.getName());
			String borderType = null;
			if (fPresentedVariableName.equals(suggestible.getName())) {
				borderType = FixBugsAnnotation.SELECTED;
			} else {
				borderType = FixBugsAnnotation.DRAGGABLE;
			}
			FixBugsAnnotation border = new FixBugsAnnotation(borderType, suggestible.getPosition());
			annotations.add(border);
		}
		setAnnotations(annotations); // update
		fLogger.info("#update(): " + (System.currentTimeMillis() - startTime) + " ms");
	}

}
