//
// Copyright (c) 2007-2012 Yoonki Song.
// All rights reserved. This program and the accompanying materials
// are made available under the permission of me.
//
// Contributors:
// Yoonki Song - initial API and implementation
//
package edu.ncsu.csc.emh.fixbugs.core.resolution.dodgyCode.catchException;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import edu.ncsu.csc.emh.fixbugs.core.util.PrintUtil;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.dodgyCode.catchException.CatchExceptionMode;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.dodgyCode.catchException.ExceptionDNDManager;

/**
 * This class is the Model class of MVC pattern.
 * 
 * @see ExceptionDNDManager (View)
 * @see CatchExceptionMode (Controller)
 * 
 * @author Yoonki Song
 */
public final class ExceptionDataModel {
	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(ExceptionDataModel.class.getName());
	static {
		fLogger.setLevel(Level.INFO);
	}

	/** K: method name, V: list of exceptions thrown by the method. */
	private List<MethodsExceptionsPair> fMethodsExceptionsToCatch;

	/** K: exception, methods may throw the exception. */
	private List<ExceptionsMethodsPair> fExceptionsMethodsToThrow;

	/** */
	private final boolean fPreserveExceptionHandler;

	/**
	 * 
	 */
	public ExceptionDataModel() {
		fMethodsExceptionsToCatch = new ArrayList<MethodsExceptionsPair>();
		fExceptionsMethodsToThrow = new ArrayList<ExceptionsMethodsPair>();
		fPreserveExceptionHandler = false;
	}

	/**
	 * 
	 * @param listToCatch
	 * @param listToThrow
	 */
	public ExceptionDataModel(final List<MethodsExceptionsPair> listToCatch, final List<ExceptionsMethodsPair> listToThrow) {
		fMethodsExceptionsToCatch = listToCatch;
		fExceptionsMethodsToThrow = listToThrow;
		fPreserveExceptionHandler = false;
	}

	/**
	 * 
	 * @param listToCatch
	 * @param listToThrow
	 * @param preserveHandler
	 */
	public ExceptionDataModel(final List<MethodsExceptionsPair> listToCatch, final List<ExceptionsMethodsPair> listToThrow,
			final boolean preserveHandler) {
		fMethodsExceptionsToCatch = listToCatch;
		fExceptionsMethodsToThrow = listToThrow;
		fPreserveExceptionHandler = preserveHandler;
	}

	public String getExceptionsByMethod(String methodName) {
		for (MethodsExceptionsPair pair : fMethodsExceptionsToCatch) {
			if (pair.getMethods().contains(methodName)) {
				return PrintUtil.toListString(pair.getExceptions());
			}
		}
		return "";
	}

	public List<String> getAllExceptions() {
		List<String> exceptions = new ArrayList<String>();
		for (MethodsExceptionsPair pair : fMethodsExceptionsToCatch) {
			exceptions.addAll(pair.getExceptions());
		}
		return exceptions;
	}

	public String getMethodsByException(String exceptionName) {
		List<String> methods = new ArrayList<String>();
		for (MethodsExceptionsPair pair : fMethodsExceptionsToCatch) {
			if (pair.getExceptions().contains(exceptionName)) {
				methods.addAll(pair.getMethods());
			}
		}
		return PrintUtil.toListString(methods);
	}

	/**
	 * 
	 * @return
	 */
	public List<MethodsExceptionsPair> getMethodsExceptionsToCatch() {
		return fMethodsExceptionsToCatch;
	}

	/**
	 * 
	 * @return
	 */
	public List<ExceptionsMethodsPair> getExceptionsMethodsToThrow() {
		return fExceptionsMethodsToThrow;
	}

	/**
	 * 
	 * @param methodsExceptionsToCatch
	 */
	public void setMethodExceptionListToCatch(List<MethodsExceptionsPair> methodsExceptionsToCatch) {
		fMethodsExceptionsToCatch = methodsExceptionsToCatch;
	}

	/**
	 * 
	 * @param exceptionsMethodsToThrow
	 */
	public void setExceptionMethodListToThrows(List<ExceptionsMethodsPair> exceptionsMethodsToThrow) {
		fExceptionsMethodsToThrow = exceptionsMethodsToThrow;
	}

	/**
	 * Example Code:
	 * 
	 * Input: E1, "Prev" of E3
	 * 
	 * <pre>
	 * m1() {
	 *   try {
	 *     ...
	 *   } catch (E1 | E2 e) {
	 *     ...
	 *   } catch ([drop] E3 e) {
	 *     ...
	 *   }
	 * }
	 * </pre>
	 * 
	 * Output:
	 * 
	 * <pre>
	 * m1() {
	 *   try {
	 *     ...
	 *   } catch (E2 e) {
	 *     ...
	 *   } catch (E1 | E3 e) {
	 *     ...
	 *   }
	 * }
	 * </pre>
	 * 
	 * @param exceptionName
	 *            the exception name being dragged
	 * @param order
	 *            "previous" or "next"
	 * @param baseExceptionName
	 *            the base exception name
	 */
	public void moveFromCaughtToCaughtOrder(final String exceptionName, final ExceptionDropPosition dropPosition) {
		fLogger.info("#moveFromCaughtToCaughtOrder: " + exceptionName + ", " + dropPosition);
		removeExceptionFromCaught(exceptionName);
		List<MethodsExceptionsPair> newMethodExceptionListToCatch = new ArrayList<MethodsExceptionsPair>();
		for (MethodsExceptionsPair pair : fMethodsExceptionsToCatch) {
			if (!pair.containsException(dropPosition.getBaseElementName())) {
				newMethodExceptionListToCatch.add(pair);
				continue;
			}
			// in most cases:
			List<String> newExceptionList = new ArrayList<String>();
			for (String existingException : pair.getExceptions()) {
				if (existingException.equals(dropPosition.getBaseElementName())) {
					if (dropPosition.isPrev()) { // previous
						newExceptionList.add(exceptionName);
						newExceptionList.add(existingException);
					} else { // next
						newExceptionList.add(existingException);
						newExceptionList.add(exceptionName);
					}
				} else {
					newExceptionList.add(existingException);
				}
			}
			MethodsExceptionsPair newPair = new MethodsExceptionsPair(pair.getMethods(), newExceptionList);
			newMethodExceptionListToCatch.add(newPair);
		}
		fMethodsExceptionsToCatch = newMethodExceptionListToCatch;
	}

	/**
	 * Input: E2
	 * 
	 * <pre>
	 * m1() {
	 *   try {
	 *     ...
	 *   } catch (E1 | E2 e) {
	 *   } [drop]
	 * }
	 * </pre>
	 * 
	 * Output:
	 * 
	 * <pre>
	 * m1() {
	 *   try {
	 *     ...
	 * } catch (E1 e) {
	 * } catch (E2 e) {
	 * }
	 * </pre>
	 * 
	 * @param exceptionName
	 *            the exception name being dragged
	 */
	public void moveFromCaughtToCaughtNoOrder(final String exceptionName) {
		fLogger.info("#moveFromCaughtToCaughtNoOrder: " + exceptionName);
		// remove
		List<String> methodListThrowException = removeExceptionFromCaught(exceptionName);
		// add it to the end of catch clause
		List<String> exceptionList = new ArrayList<String>();
		exceptionList.add(exceptionName);
		MethodsExceptionsPair pairToAdd = new MethodsExceptionsPair(methodListThrowException, exceptionList);
		fMethodsExceptionsToCatch.add(pairToAdd);
	}

	/**
	 * Input: E2, "0"th index of "catch" clauses.
	 * 
	 * <pre>
	 * m1() {
	 *   try {
	 *     ...
	 *   } [drop] catch (E1 | E2 e) {
	 *   }
	 * }
	 * </pre>
	 * 
	 * Output:
	 * 
	 * <pre>
	 * m1() {
	 *   try {
	 *     ...
	 *   } catch (E2 e) {
	 *   } catch (E1 e) {
	 *   }
	 * }
	 * </pre>
	 * <p>
	 * Adds a new uni-catch to the given index of CatchClauses.
	 * 
	 * @param exceptionName
	 *            the exception name being dragged
	 */
	public void moveFromMultiCaughtToCatch(final String exceptionName, final ExceptionDropPosition dropPosition) {
		fLogger.info("#moveFromMultiCaughtToCatch: " + exceptionName + ", " + dropPosition.getCatchClauseIndex());

		// remove the exception being dragged.
		List<String> methodListThrowException = removeExceptionFromCaught(exceptionName);
		List<String> exceptionList = new ArrayList<String>();
		exceptionList.add(exceptionName);
		MethodsExceptionsPair newPair = new MethodsExceptionsPair(methodListThrowException, exceptionList);

		// add the new pair to the beginning or the middle of CatchClauses
		List<MethodsExceptionsPair> newMethodsExceptionsToCatch = new ArrayList<MethodsExceptionsPair>();
		for (int i = 0; i < fMethodsExceptionsToCatch.size(); i++) {
			if (i == dropPosition.getCatchClauseIndex()) {
				// add it to the target index of catch clauses
				newMethodsExceptionsToCatch.add(newPair);
			}
			newMethodsExceptionsToCatch.add(fMethodsExceptionsToCatch.get(i));
		}
		// or, add it to the end of CatchClauses.
		if (dropPosition.getCatchClauseIndex() == fMethodsExceptionsToCatch.size()) {
			newMethodsExceptionsToCatch.add(newPair);
		}

		fMethodsExceptionsToCatch = newMethodsExceptionsToCatch;
	}

	/**
	 * Input: E1, "2"th index of "catch" clauses.
	 * 
	 * <pre>
	 * m1() {
	 *   try {
	 *     ...
	 *   } catch (E1 e) {
	 *   } catch (E2 e) {
	 *   } [drop]
	 * }
	 * </pre>
	 * 
	 * Output:
	 * 
	 * <pre>
	 * m1() {
	 *   try {
	 *     ...
	 *   } catch (E2 e) {
	 *   } catch (E1 e) {
	 *   }
	 * }
	 * </pre>
	 * 
	 * <p>
	 * Changes the order of CatchClauses.
	 * 
	 * @param exceptionName
	 *            the exception name being dragged
	 */
	public void moveFromCaughtToCatch(final String exceptionName, final ExceptionDropPosition dropPosition) {
		fLogger.info("#moveFromCaughtToCatch: " + exceptionName + ", " + dropPosition.getCatchClauseIndex());

		// remove the exception being dragged.
		List<String> methodListThrowException = removeExceptionFromCaught(exceptionName);
		List<String> exceptionList = new ArrayList<String>();
		exceptionList.add(exceptionName);
		MethodsExceptionsPair newPair = new MethodsExceptionsPair(methodListThrowException, exceptionList);

		// add the new pair to the beginning or the middle of CatchClauses
		List<MethodsExceptionsPair> newMethodsExceptionsToCatch = new ArrayList<MethodsExceptionsPair>();
		for (int i = 0; i < fMethodsExceptionsToCatch.size(); i++) {
			if (i == dropPosition.getCatchClauseIndex()) {
				// add it to the target index of catch clauses
				newMethodsExceptionsToCatch.add(newPair);
			}
			newMethodsExceptionsToCatch.add(fMethodsExceptionsToCatch.get(i));
		}
		// or, add it to the end of CatchClauses.
		if (dropPosition.getCatchClauseIndex() - 1 == fMethodsExceptionsToCatch.size()) {
			newMethodsExceptionsToCatch.add(newPair);
		}

		fMethodsExceptionsToCatch = newMethodsExceptionsToCatch;
	}

	/**
	 * <p>
	 * Input: E2, "Next" of E1
	 * 
	 * <pre>
	 * m1() throws E1 [drop] {
	 *   try {
	 *     ...
	 *   } catch (E2 | E3 e) {
	 *   }
	 * }
	 * </pre>
	 * 
	 * <p>
	 * Output:
	 * 
	 * <pre>
	 * m1() throws E1, E2 {
	 *   try {
	 *     ...
	 *   } catch (E3 e) {
	 *   }
	 * }
	 * </pre>
	 * 
	 * @param exceptionName
	 *            the exception name being dragged
	 * @param order
	 *            "previous" or "next"
	 * @param baseExceptionName
	 *            the base exception name
	 */
	public void moveFromCaughtToThrownExcpeptions(final String exceptionName, final ExceptionDropPosition dropPosition) {
		fLogger.info("#moveFromCaughtToThrownOrder: " + exceptionName + ", " + dropPosition);
		// remove
		List<String> methodListThrowException = removeExceptionFromCaught(exceptionName);

		// add
		List<ExceptionsMethodsPair> newExceptionMethodListToThrows = new ArrayList<ExceptionsMethodsPair>();
		for (ExceptionsMethodsPair pair : fExceptionsMethodsToThrow) {
			// we found the base exception
			if (pair.containsException(dropPosition.getBaseElementName())) {
				List<String> exceptionList = new ArrayList<String>();
				exceptionList.add(exceptionName);
				ExceptionsMethodsPair pairToAdd = new ExceptionsMethodsPair(exceptionList, methodListThrowException);
				fLogger.info("\tAdd: " + pairToAdd);
				if (dropPosition.isPrev()) { // Previous
					newExceptionMethodListToThrows.add(pairToAdd);
					newExceptionMethodListToThrows.add(pair);
				} else { // Next
					newExceptionMethodListToThrows.add(pair);
					newExceptionMethodListToThrows.add(pairToAdd);
				}
			} else {
				newExceptionMethodListToThrows.add(pair);
			}
		}

		fExceptionsMethodsToThrow = newExceptionMethodListToThrows;
	}

	/**
	 * <p>
	 * Input: E1,
	 * 
	 * <pre>
	 * m1() [drop] {
	 *   try {
	 *     ...
	 *   } catch (E1 | E2 e) {
	 *   }
	 * }
	 * </pre>
	 * 
	 * <p>
	 * Output:
	 * 
	 * <pre>
	 * m1() throws E1 {
	 *   try {
	 *     ...
	 *   } catch (E2 e) {
	 *   }
	 * }
	 * </pre>
	 * 
	 * @param exceptionName
	 *            the exception name being dragged
	 */
	public void moveFromCaughtToEmptyThrows(final String exceptionName) {
		fLogger.info("#moveFromCaughtToThrownNoOrder: " + exceptionName);
		// remove
		List<String> methodListThrowException = removeExceptionFromCaught(exceptionName);
		// add
		List<String> exceptionList = new ArrayList<String>();
		exceptionList.add(exceptionName);
		fExceptionsMethodsToThrow.add(new ExceptionsMethodsPair(exceptionList, methodListThrowException));
	}

	/**
	 * Example Code:
	 * 
	 * <pre>
	 * m1() throws E1 {
	 *   try {
	 *     ...
	 *   } catch (E2 e) {
	 *   }
	 * }
	 * </pre>
	 * 
	 * Input1: E1, "Prev" of E2
	 * <p>
	 * Output1:
	 * 
	 * <pre>
	 * m1() {
	 *   try {
	 *     ...
	 *   } catch (E1 | E2 e) {
	 *   }
	 * }
	 * </pre>
	 * 
	 * Input2: E1, "Next" of E2
	 * <p>
	 * Output2:
	 * 
	 * <pre>
	 * m1() {
	 *   try {
	 *     ...
	 *   } catch (E2 | E1 e) {
	 *   }
	 * }
	 * </pre>
	 * 
	 * @param exceptionName
	 *            the exception name being dragged
	 * @param order
	 *            "previous" or "next"
	 * @param baseExceptionName
	 *            the base exception name
	 */
	public void moveFromThrownToCaughtOrder(final String exceptionName, final ExceptionDropPosition dropPosition) {
		fLogger.info("#moveFromThrownToCaughtOrder: " + exceptionName + ", " + dropPosition);
		// remove it from the exceptionsToCatch
		removeExceptionFromThrown(exceptionName);
		// add it to the methodExceptionListToCatch
		List<MethodsExceptionsPair> newMethodExceptionListToCatch = new ArrayList<MethodsExceptionsPair>();
		for (MethodsExceptionsPair pair : fMethodsExceptionsToCatch) {
			if (!pair.containsException(dropPosition.getBaseElementName())) {
				newMethodExceptionListToCatch.add(pair);
				continue;
			}
			// in most cases:
			List<String> newExceptionList = new ArrayList<String>();
			for (String existingException : pair.getExceptions()) {
				if (existingException.equals(dropPosition.getBaseElementName())) {
					if (dropPosition.isPrev()) { // Prev
						newExceptionList.add(exceptionName);
						newExceptionList.add(existingException);
					} else { // Next
						newExceptionList.add(existingException);
						newExceptionList.add(exceptionName);
					}
				} else {
					newExceptionList.add(existingException);
				}
			}
			MethodsExceptionsPair newPair = new MethodsExceptionsPair(pair.getMethods(), newExceptionList);
			newMethodExceptionListToCatch.add(newPair);
		}
		fMethodsExceptionsToCatch = newMethodExceptionListToCatch;
	}

	/**
	 * Input: E1, 1st index of CatchClauses.
	 * 
	 * <pre>
	 * m1() throws E1 {
	 *   try {
	 *     ...
	 *   } catch (E2 e) {
	 *   } [drop]
	 * }
	 * </pre>
	 * 
	 * <p>
	 * Output:
	 * 
	 * <pre>
	 * m1() {
	 *   try {
	 *     ...
	 *   } catch (E2 e) {
	 *   } catch (E1 e) {
	 *   }
	 * }
	 * </pre>
	 * 
	 * Adds a new uni-catch to the given index of CatchClauses.
	 * 
	 * @param exceptionName
	 * @param dropPosition
	 */
	public void moveFromThrownToCatch(final String exceptionName, final ExceptionDropPosition dropPosition) {
		fLogger.info("#moveFromThrownToCatch: " + exceptionName + ", " + dropPosition);
		// remove it from the exceptionsToCatch
		List<String> methodListThrowException = removeExceptionFromThrown(exceptionName);
		List<String> exceptionList = new ArrayList<String>();
		exceptionList.add(exceptionName);
		MethodsExceptionsPair pairToAdd = new MethodsExceptionsPair(methodListThrowException, exceptionList);

		// add it to the beginning and the middle of CatchClause
		List<MethodsExceptionsPair> newMethodsExceptionsToCatch = new ArrayList<MethodsExceptionsPair>();
		for (int i = 0; i < fMethodsExceptionsToCatch.size(); i++) {
			if (i == dropPosition.getCatchClauseIndex()) {
				// add it to the target index of catch clauses
				newMethodsExceptionsToCatch.add(pairToAdd);
			}
			newMethodsExceptionsToCatch.add(fMethodsExceptionsToCatch.get(i));
		}
		// or, add it to the end of CatchClause.
		if (dropPosition.getCatchClauseIndex() == fMethodsExceptionsToCatch.size()) {
			newMethodsExceptionsToCatch.add(pairToAdd);
		}
		fMethodsExceptionsToCatch = newMethodsExceptionsToCatch;
	}

	/**
	 * Input: E1, end(1) of the CatchClause containing 'E2'.
	 * 
	 * <pre>
	 * m1() throws E1 {
	 *   try {
	 *     ...
	 *   } catch (E2 e) {
	 *   } [drop]
	 * }
	 * </pre>
	 * 
	 * <pre>
	 * m1() {
	 *   try {
	 *     ...
	 *   } catch (E2 e) {
	 *   } catch (E1 e) {
	 *   }
	 * }
	 * </pre>
	 * 
	 * @param exceptionName
	 *            the exception name being dragged
	 */
	public void moveFromThrownToCaughtNoOrder(final String exceptionName) {
		fLogger.info("#moveFromThrownToCaughtNoOrder: " + exceptionName);
		// remove it from the exceptionsToCatch
		List<String> methodListThrowException = removeExceptionFromThrown(exceptionName);

		// add it to the methodExceptionListToCatch
		List<String> exceptionListThrow = new ArrayList<String>();
		exceptionListThrow.add(exceptionName);
		MethodsExceptionsPair newPair = new MethodsExceptionsPair(methodListThrowException, exceptionListThrow);
		fMethodsExceptionsToCatch.add(newPair);
	}

	/**
	 * Input: E1, end(1) of the CatchClause containing 'E2'.
	 * 
	 * <pre>
	 * m1() throws E1 {
	 *   try {
	 *     ...
	 *   } catch (E2 e) {
	 *   } [drop]
	 * }
	 * </pre>
	 * 
	 * <pre>
	 * m1() {
	 *   try {
	 *     ...
	 *   } catch (E2 e) {
	 *   } catch (E1 e) {
	 *   }
	 * }
	 * </pre>
	 * 
	 * @param exceptionName
	 *            the exception name being dragged
	 */
	public void moveFromThrownToCatchNoOrder(final String exceptionName) {
		fLogger.info("#moveFromThrownToCatchNoOrder: " + exceptionName);
		// remove it from the exceptionsToCatch
		List<String> methodListThrowException = removeExceptionFromThrown(exceptionName);

		// add it to the methodExceptionListToCatch
		List<String> exceptionListThrow = new ArrayList<String>();
		exceptionListThrow.add(exceptionName);
		MethodsExceptionsPair newPair = new MethodsExceptionsPair(methodListThrowException, exceptionListThrow);
		fMethodsExceptionsToCatch.add(newPair);
	}

	/**
	 * <p>
	 * Input: E1, "Next" of E2
	 * 
	 * <pre>
	 * m1() throws E1, E2 {
	 *     ...
	 * } catch (E3 e) {
	 *     ...
	 * }
	 * </pre>
	 * 
	 * Output:
	 * 
	 * <pre>
	 * m1() throws E2, E1 {
	 *     ...
	 * } catch (E3 e) {
	 *     ...
	 * }
	 * </pre>
	 * 
	 * @param exceptionName
	 *            the exception name being dragged
	 * @param order
	 *            "previous" or "next"
	 * @param baseExceptionName
	 *            the base exception name
	 */
	public void moveFromThrownToThrownOrder(final String exceptionName, final ExceptionDropPosition dropPosition) {
		fLogger.info("#moveFromThrownToThrownOrder: " + exceptionName + ", " + dropPosition);
		List<String> methodListThrowException = removeExceptionFromThrown(exceptionName);
		// add
		List<ExceptionsMethodsPair> newExceptionMethodListToThrows = new ArrayList<ExceptionsMethodsPair>();
		for (ExceptionsMethodsPair pair : fExceptionsMethodsToThrow) {
			// ignore itself
			if (pair.containsException(exceptionName)) {
				continue;
			}
			if (pair.containsException(dropPosition.getBaseElementName())) {
				List<String> exceptionList = new ArrayList<String>();
				exceptionList.add(exceptionName);
				ExceptionsMethodsPair pairToAdd = new ExceptionsMethodsPair(exceptionList, methodListThrowException);
				fLogger.info("\tAdd: " + pairToAdd);
				if (dropPosition.isPrev()) { // Previous
					newExceptionMethodListToThrows.add(pairToAdd);
					newExceptionMethodListToThrows.add(pair);
				} else { // Next
					newExceptionMethodListToThrows.add(pair);
					newExceptionMethodListToThrows.add(pairToAdd);
				}
			} else {
				newExceptionMethodListToThrows.add(pair);
			}
		}
		fExceptionsMethodsToThrow = newExceptionMethodListToThrows;
	}

	private List<String> removeExceptionFromThrown(final String exceptionName) {
		List<String> methodListThrowException = new ArrayList<String>();
		List<ExceptionsMethodsPair> newExceptionMethodListToThrows = new ArrayList<ExceptionsMethodsPair>();
		for (ExceptionsMethodsPair pair : fExceptionsMethodsToThrow) {

			// if the list contains a single exception, then skip the list!
			if (!pair.containsException(exceptionName)) {
				newExceptionMethodListToThrows.add(pair);
			} else {
				methodListThrowException = pair.getMethods();
			}
		}
		fExceptionsMethodsToThrow = newExceptionMethodListToThrows;
		return methodListThrowException;
	}

	private List<String> removeExceptionFromCaught(final String exceptionToRemove) {
		fLogger.info("#removeExceptionFromCaught: " + exceptionToRemove);
		List<String> methodListThrowException = new ArrayList<String>();
		List<MethodsExceptionsPair> newMethodExceptionListToCatch = new ArrayList<MethodsExceptionsPair>();
		for (MethodsExceptionsPair pair : fMethodsExceptionsToCatch) {
			List<String> newExceptionList = new ArrayList<String>();

			// if the list contains a single exception, then skip the list!
			if (pair.getExceptions().size() == 1) {
				if (pair.containsException(exceptionToRemove)) {
					// method name
					methodListThrowException.addAll(pair.getMethods());
					continue;
				}
			}

			for (String exceptionName : pair.getExceptions()) {
				if (!exceptionName.contains(exceptionToRemove)) {
					newExceptionList.add(exceptionName);
				} else {
					// method name
					methodListThrowException.addAll(pair.getMethods());
				}
			}
			newMethodExceptionListToCatch.add(new MethodsExceptionsPair(pair.getMethods(), newExceptionList));
		}
		fMethodsExceptionsToCatch = newMethodExceptionListToCatch;
		return methodListThrowException;
	}

	/**
	 * It's for specific purpose only.
	 * 
	 * <p>
	 * Input:
	 * 
	 * <pre>
	 * [[m1, [e1, e2, e3]], e2]
	 * </pre>
	 * 
	 * Output:
	 * 
	 * <pre>
	 * [[m1, [e1, e3]], [m1, [e2]]]
	 * </pre>
	 * 
	 * @param list
	 * @param element
	 * @return
	 */
	private static List<List<String>> splitExceptionList(List<String> list, final String element) {
		List<List<String>> listOfList = new ArrayList<List<String>>();
		list.remove(element);
		listOfList.add(list);
		listOfList.add(Arrays.asList(element));
		return listOfList;
	}

	/**
	 * it's for specific purpose only.
	 * 
	 * <p>
	 * Input: "E1", "Before" of "E3"
	 * 
	 * <pre>
	 * [[m1, [E1, E2]], [m2, [E3]]]
	 * </pre>
	 * 
	 * <p>
	 * Output:
	 * 
	 * <pre>
	 * [[m1, [E2]], [m1, [E1, E3]]]
	 * </pre>
	 * 
	 * @param list
	 * @param element
	 * @return
	 */
	private static List<List<String>> joinExceptionList(final List<String> list, final String element) {
		List<List<String>> listOfList = new ArrayList<List<String>>();
		list.remove(element);
		listOfList.add(list);
		listOfList.add(Arrays.asList(element));
		return listOfList;
	}

	public boolean isMultiCaught(String exceptionName) {
		for (MethodsExceptionsPair pair : fMethodsExceptionsToCatch) {
			if (pair.getExceptions().size() >= 2 && pair.containsException(exceptionName)) {
				return true;
			}
		}
		return false;
	}

	@Override
	public ExceptionDataModel clone() throws CloneNotSupportedException {
		List<MethodsExceptionsPair> newMethodsExceptionsPair = new ArrayList<MethodsExceptionsPair>();
		for (MethodsExceptionsPair pair : getMethodsExceptionsToCatch()) {
			List<String> newMethods = new ArrayList<String>();
			for (String method : pair.getMethods()) {
				newMethods.add(method);
			}
			List<String> newExceptions = new ArrayList<String>();
			for (String exception : pair.getExceptions()) {
				newExceptions.add(exception);
			}
			MethodsExceptionsPair newPair = new MethodsExceptionsPair(newMethods, newExceptions);
			newMethodsExceptionsPair.add(newPair);
		}
		List<ExceptionsMethodsPair> newExceptionsMethodsPair = new ArrayList<ExceptionsMethodsPair>();
		for (ExceptionsMethodsPair pair : getExceptionsMethodsToThrow()) {
			List<String> newExceptions = new ArrayList<String>();
			for (String exception : pair.getExceptions()) {
				newExceptions.add(exception);
			}
			List<String> newMethods = new ArrayList<String>();
			for (String method : pair.getMethods()) {
				newMethods.add(method);
			}
			ExceptionsMethodsPair newPair = new ExceptionsMethodsPair(newExceptions, newMethods);
			newExceptionsMethodsPair.add(newPair);
		}

		return new ExceptionDataModel(newMethodsExceptionsPair, newExceptionsMethodsPair);
	}

	@Override
	public String toString() {
		StringBuilder buf = new StringBuilder();
		buf.append("------------------------------------------\n");
		buf.append("Exceptions to catch:\n");
		for (MethodsExceptionsPair pair : fMethodsExceptionsToCatch) {
			buf.append(pair.toString()).append('\n');
		}

		buf.append("\nExceptions to throw:\n");
		for (ExceptionsMethodsPair pair : fExceptionsMethodsToThrow) {
			buf.append(pair.toString()).append('\n');
		}
		buf.append("------------------------------------------\n");
		return buf.toString();
	}
}
