//
// Copyright (c) 2007-2012 Yoonki Song.
// All rights reserved. This program and the accompanying materials
// are made available under the permission of me.
//
// Contributors:
// Yoonki Song - initial API and implementation
//
package edu.ncsu.csc.emh.fixbugs.core.resolution.dodgyCode.catchException;

import org.eclipse.jface.text.TypedPosition;

/**
 * @author Yoonki Song
 */
public final class ExceptionDropPosition extends TypedPosition {

	/** */
	public static final String IN_CATCH_CLAUSE = "in_catch_clause";

	/** */
	public static final String OUT_OF_CATCH_CLAUSE = "out_of_catch_clause";

	/** */
	public static final String IN_THROWN_EXCEPTIONS = "in_thrown_exceptions";

	/** */
	public static final String IN_EMPTY_THROWS = "in_empty_throws";

	/** */
	private final String fPlace;

	/** the position before the given target. */
	public static final String PREV = "prev";

	/** the position after the given target. */
	public static final String NEXT = "next";

	/** the position of the given target. */
	public static final String ITSELF = "itself";

	/** */
	public static final String UNKNOWN_NAME = "unknown_name";

	/** */
	private final String fBaseElementName;

	/**
	 * index of CatchClause. It is only useful when the drop Zone Kind is a
	 * CATCH_CLAUSE.
	 */
	private int fCatchClauseIndex = 0;

	/**
	 * Constructor for DropPosition.
	 * 
	 * @param offset
	 *            the offset
	 * @param length
	 *            the length
	 * @param type
	 *            the position is in throws or catch.
	 * @param place
	 *            the previous of the next of the given exception
	 */
	public ExceptionDropPosition(final int offset, final int length, final String type, final String place) {
		this(offset, length, type, place, UNKNOWN_NAME);
	}

	public ExceptionDropPosition(final int offset, final int length, final String type, final String place,
			final String baseElementName) {
		this(offset, length, type, place, baseElementName, 0);
	}

	public ExceptionDropPosition(final int offset, final int length, final String type, final int index) {
		this(offset, length, type, ITSELF, UNKNOWN_NAME, index);
	}

	/**
	 * Constructor for DropPosition.
	 * 
	 * @param offset
	 *            the offset
	 * @param length
	 *            the length
	 * @param type
	 *            the position is in throws or catch.
	 * @param place
	 *            the previous of the next of the given exception
	 * @param baseElementName
	 *            the base element name (exception)
	 */
	public ExceptionDropPosition(final int offset, final int length, final String type, final String place,
			final String baseElementName, final int index) {
		super(offset, length, type);
		fPlace = place;
		fBaseElementName = baseElementName;
		fCatchClauseIndex = index;
	}

	/**
	 * Returns the base element name.
	 * 
	 * @return the base element name.
	 */
	public String getBaseElementName() {
		return fBaseElementName;
	}

	/**
	 * Returns the place.
	 * 
	 * @return the place.
	 */
	public String getPlace() {
		return fPlace;
	}

	/**
	 * Checks if the position is in catch clause.
	 * 
	 * @return true if the position is in catch clause; false otherwise.
	 */
	public boolean isInCatchClause() {
		return (IN_CATCH_CLAUSE.equals(getType()));
	}

	/**
	 * Checks if the position is in catch clause.
	 * 
	 * @return true if the position is in catch clause; false otherwise.
	 */
	public boolean isOutOfCatchClause() {
		return (OUT_OF_CATCH_CLAUSE.equals(getType()));
	}

	/**
	 * Checks if the position is in throws clause.
	 * 
	 * @return true if the position is in throws clause; false otherwise.
	 */
	public boolean isInThrownExceptions() {
		return (IN_THROWN_EXCEPTIONS.equals(getType()));
	}

	/**
	 * Checks if the position is in throws clause.
	 * 
	 * @return true if the position is in throws clause; false otherwise.
	 */
	public boolean isInEmptyThrows() {
		return (IN_EMPTY_THROWS.equals(getType()));
	}

	/**
	 * Checks if the position is the previous position of an exception.
	 * 
	 * @return true if the position is the previous position of an exception;
	 *         false otherwise.
	 */
	public boolean isPrev() {
		return (PREV.equals(fPlace));
	}

	/**
	 * Checks if the position is the next position of an exception.
	 * 
	 * @return true if the position is the next position of an exception; false
	 *         otherwise.
	 */
	public boolean isNext() {
		return (NEXT.equals(fPlace));
	}

	/**
	 * Checks if there is no need to care about the position of an exception.
	 * 
	 * @return true if there is no need to care about the position of an
	 *         exception; false otherwise.
	 */
	public boolean isItself() {
		return (ITSELF.equals(fPlace));
	}

	public int getCatchClauseIndex() {
		return fCatchClauseIndex;
	}

	@Override
	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append("ThrowsOrCatch: " + getType());
		buf.append(", Place: " + fPlace);
		buf.append(", BaseElementName: " + fBaseElementName);
		buf.append(", Offset: " + getOffset());
		buf.append(", Length: " + getLength());
		return buf.toString();
	}
}
