//
// Copyright (c) 2007-2012 Yoonki Song.
// All rights reserved. This program and the accompanying materials
// are made available under the permission of me.
//
// Contributors:
// Yoonki Song - initial API and implementation
//
package edu.ncsu.csc.emh.fixbugs.core.resolution.dodgyCode.catchException;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.BodyDeclaration;
import org.eclipse.jdt.core.dom.CatchClause;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.TryStatement;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.UnionType;
import org.eclipse.jdt.internal.ui.text.correction.ASTResolving;
import org.eclipse.jface.text.Position;

import edu.ncsu.csc.emh.fixbugs.core.annotation.ExceptionAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.annotation.FixBugsAnnotation;

/**
 * 
 * @author Yoonki Song
 */
public final class ExceptionDropPositionFinder {
	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(ExceptionDropPositionFinder.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	/**
	 * No instance.
	 */
	private ExceptionDropPositionFinder() {
	}

	/**
	 * Finds drop positions for annotations being dragged.
	 * 
	 * @param assistant
	 *            the current assistant.
	 * @param position
	 *            the exception position.
	 * @param name
	 *            the exception name.
	 * @param isThrown
	 *            true if the source exception is thrown; false otherwise.
	 * @return the array of drop positions.
	 */
	public static ExceptionDropPosition[] perform(final CatchExceptionAssistant assistant, final Position position,
			final String name, final boolean isThrown) {
		fLogger.info("ExceptionDropPositionFinder#perform: ");
		SimpleName exceptionName = (SimpleName) assistant.findASTNode(position);
		List<ExceptionDropPosition> positions = new ArrayList<ExceptionDropPosition>();
		// find drop zones
		if (isCaught(exceptionName)) { // from
			if (isMultiCaught(exceptionName)) {
				// multi-catch
				findPosForMultiCaughtException(positions, assistant, exceptionName);
			} else {
				// uni-catch
				findPosForUniCatchException(positions, assistant, exceptionName);
			}
		} else if (isThrown(exceptionName)) {
			findPosForThrownException(positions, assistant, exceptionName);
		}

		// add itself
		Position p = position;
		String throwOrCatch = null;
		if (isThrown) { // if (annotation.isThrown()) {
			throwOrCatch = ExceptionDropPosition.IN_THROWN_EXCEPTIONS;
		} else {
			throwOrCatch = ExceptionDropPosition.IN_CATCH_CLAUSE;
		}
		ExceptionDropPosition dragPosition = new ExceptionDropPosition(p.getOffset(), p.getLength(), throwOrCatch,
				ExceptionDropPosition.ITSELF);
		positions.add(dragPosition);
		return positions.toArray(new ExceptionDropPosition[] {});
	}

	/**
	 * finds drop positions for an exception in multi-catch.
	 * 
	 * @param positions
	 * @param assistant
	 * @param node
	 * @return
	 */
	private static List<ExceptionDropPosition> findPosForMultiCaughtException(final List<ExceptionDropPosition> positions,
			final CatchExceptionAssistant assistant, final ASTNode node) {
		fLogger.info("ExceptionDropPositionFinder#findPosForMultiCaughtException");

		// find positions in throws clause
		findInThrows(positions, assistant, node);

		// find prev and/or next positions of exceptions in catch
		// clauses selectively
		findInCatchClauses(positions, assistant, node);

		// find the beginning or the end of CatchClauses.
		findOutOfCatchClause(positions, assistant, node);

		return positions;
	}

	private static List<ExceptionDropPosition> findPosForUniCatchException(final List<ExceptionDropPosition> positions,
			final CatchExceptionAssistant assistant, final ASTNode node) {
		fLogger.info("ExceptionDropPositionFinder#findPosForUniCatchException");

		if (getCaughtExceptionsSize(assistant) == 1 && !hasFinallyBlock(node)) {
			return positions; // do nothing!
		}

		// add positions in throws clause
		findInThrows(positions, assistant, node);

		// add the beginning and the end position of exceptions in catch clauses
		// selectively
		findInCatchClauses(positions, assistant, node);

		// find the beginning or the end of CatchClauses.
		findOutOfCatchClause(positions, assistant, node);

		return positions;
	}

	private static int getCaughtExceptionsSize(final CatchExceptionAssistant assistant) {
		int length = 0;
		for (FixBugsAnnotation a : assistant.getAnnotations()) {
			if (!(a instanceof ExceptionAnnotation)) {
				continue;
			}
			ExceptionAnnotation ea = (ExceptionAnnotation) a;
			if (ea.isCaught()) {
				length++;
			}
		}
		return length;
	}

	private static boolean hasFinallyBlock(final ASTNode node) {
		TryStatement tryStatement = ASTResolving.findParentTryStatement(node);
		return tryStatement.getFinally() != null;
	}

	// Finds drop positions for a thrown exception.
	private static List<ExceptionDropPosition> findPosForThrownException(final List<ExceptionDropPosition> positions,
			final CatchExceptionAssistant assistant, final ASTNode node) {
		fLogger.info("ExceptionDropPositionFinder#findPosForThrownException");

		// add the begin position of method declaration
		findInThrows(positions, assistant, node);

		// add the begin and end position of exceptions in catch clauses
		// selectively
		findInCatchClauses(positions, assistant, node);

		// add the end position of "try" block
		// findEndPosOfTryStatement2(positions, assistant);

		// find the beginning or the end of CatchClauses.
		findOutOfCatchClause(positions, assistant, node);

		return positions;
	}

	private static List<ExceptionDropPosition> findInThrownExceptions(final List<ExceptionDropPosition> positions,
			final CatchExceptionAssistant assistant, final ASTNode node) {
		fLogger.info("ExceptionDropPositionFinder#findPrevNextPosInThrownExceptions");
		for (FixBugsAnnotation a : assistant.getAnnotations()) {
			// ignore other annotations
			if (!(a instanceof ExceptionAnnotation)) {
				continue;
			}
			// ignore exception annotations in throws, or itself
			ExceptionAnnotation ea = (ExceptionAnnotation) a;
			if (!ea.isThrown() || a.includes(node.getStartPosition())) {
				continue;
			}

			Position p = a.getPosition();
			SimpleName targetNode = (SimpleName) assistant.findASTNode(a.getPosition());
			int targetOrder = findOrderOfThrownException(targetNode);

			// add the prev position of an Exception
			if (targetOrder == 0) {
				positions.add(new ExceptionDropPosition(p.getOffset() - 1, 1, ExceptionDropPosition.IN_THROWN_EXCEPTIONS,
						ExceptionDropPosition.PREV, a.getName()));
			}
			// add the next position of an Exception
			positions.add(new ExceptionDropPosition(p.getOffset() + p.getLength(), 1,
					ExceptionDropPosition.IN_THROWN_EXCEPTIONS, ExceptionDropPosition.NEXT, a.getName())); // THROWS??
		}
		return positions;
	}

	private static List<ExceptionDropPosition> findInCatchClauses(final List<ExceptionDropPosition> positions,
			final CatchExceptionAssistant assistant, final ASTNode node) {
		fLogger.info("ExceptionDropPositionFinder#findInCaughtExceptions");
		for (FixBugsAnnotation a : assistant.getAnnotations()) {
			// ignore other annotations
			if (!(a instanceof ExceptionAnnotation)) {
				continue;
			}
			// ignore exception annotations in throws, or itself
			ExceptionAnnotation ea = (ExceptionAnnotation) a;
			if (ea.isThrown() || a.includes(node.getStartPosition())) {
				continue;
			}

			Position p = a.getPosition();
			ExceptionDropPosition prev = new ExceptionDropPosition(p.getOffset() - 1, 1, ExceptionDropPosition.IN_CATCH_CLAUSE,
					ExceptionDropPosition.PREV, a.getName());
			ExceptionDropPosition next = new ExceptionDropPosition(p.getOffset() + p.getLength(), 1,
					ExceptionDropPosition.IN_CATCH_CLAUSE, ExceptionDropPosition.NEXT, a.getName());
			if (ea.isMultiCaught()) {
				// get target ...
				SimpleName targetNode = (SimpleName) assistant.findASTNode(a.getPosition());
				int targetOrder = findOrderOfMultiCaughtExceptions(targetNode);
				if (isInSameMultiCatch(targetNode, (SimpleName) node)) {
					// care about order
					int sourceOrder = findOrderOfMultiCaughtExceptions((SimpleName) node);
					fLogger.info("order: " + sourceOrder);
					if (sourceOrder > targetOrder) {
						// add the previous position of Exception
						positions.add(prev);
					} else if (sourceOrder == targetOrder) {
						// do nothing
					} else {
						// add the next position of Exception
						positions.add(next); // THROWS??
					}
				} else { // different multi-catch
					if (targetOrder == 0) {
						// add the previous position of Exception
						positions.add(prev);
					}
					// add the next position of Exception
					positions.add(next); // THROWS??
				}
			} else if (ea.isCaught()) {
				// add the previous position of Exception
				positions.add(prev);
				// add the next position of Exception
				positions.add(next); // THROWS??
			}
		}
		return positions;
	}

	private static boolean isInSameMultiCatch(final SimpleName exception1, final SimpleName exception2) {
		if (isMultiCaught(exception1) && isMultiCaught(exception2)) {
			UnionType ut1 = (UnionType) exception1.getParent().getParent();
			UnionType ut2 = (UnionType) exception2.getParent().getParent();
			if (ut1.equals(ut2)) {
				return true;
			}
		}
		return false;
	}

	private static void findInThrows(final List<ExceptionDropPosition> positions, final CatchExceptionAssistant assistant,
			final ASTNode node) {
		BodyDeclaration bodyDeclaration = ASTResolving.findParentBodyDeclaration(node);
		fLogger.info("ExceptionDropPositionFinder#findInThrownExceptions");
		if (bodyDeclaration instanceof MethodDeclaration) {
			// the beginning of method body
			MethodDeclaration md = (MethodDeclaration) bodyDeclaration;
			if (md.thrownExceptions().size() == 0) {
				// add the begin of method body
				findInEmptyThrows(positions, md, node);
			} else {
				// add the previous and next of exceptions in throws
				findInThrownExceptions(positions, assistant, node);
			}
		}
	}

	// add the begin of method body
	private static void findInEmptyThrows(final List<ExceptionDropPosition> positions, final MethodDeclaration md,
			final ASTNode node) {
		fLogger.info("ExceptionDropPositionFinder#findBeginOfMethodBodyDeclaration");
		Block block = md.getBody();
		positions.add(new ExceptionDropPosition(block.getStartPosition() - 1, 1, ExceptionDropPosition.IN_EMPTY_THROWS,
				ExceptionDropPosition.ITSELF));
	}

	/**
	 * The given node is in catch clause.
	 * 
	 * @param positions
	 * @param node
	 */
	// private static void findEndPosOfTryBlock(final
	// List<ExceptionDropPosition> positions, final ASTNode node) {
	// fLogger.info("ExceptionDropPositionFinder#findEndPositionOfTryBlock");
	// TryStatement tryStatement = ASTResolving.findParentTryStatement(node);
	// if (tryStatement != null) {
	// Block tryBlock = tryStatement.getBody();
	// positions.add(new ExceptionDropPosition(tryBlock.getStartPosition() +
	// tryBlock.getLength(), 1,
	// ExceptionDropPosition.KIND_CAUGHT_EXCEPTION,
	// ExceptionDropPosition.ITSELF));
	// }
	// }

	private static void findOutOfCatchClause(final List<ExceptionDropPosition> positions,
			final CatchExceptionAssistant assistant, final ASTNode node) {
		fLogger.info("ExceptionDropPositionFinder#findInCatchClause");
		TryStatement tryStatement = assistant.findTryStatement();
		if (tryStatement != null) {
			// no catch, only finally block exists.
			if (tryStatement.catchClauses().isEmpty()) {
				Block tryBlock = tryStatement.getBody();
				positions.add(new ExceptionDropPosition(tryBlock.getStartPosition() + tryBlock.getLength(), 1,
						ExceptionDropPosition.OUT_OF_CATCH_CLAUSE, 0));
			} else {
				int size = tryStatement.catchClauses().size();
				for (int i = 0; i < size; i++) {
					CatchClause cc = (CatchClause) tryStatement.catchClauses().get(i);
					// add the beginning point of each CatchClause
					positions.add(new ExceptionDropPosition(cc.getStartPosition() - 1, 1,
							ExceptionDropPosition.OUT_OF_CATCH_CLAUSE, i));
					if (i == size - 1) { // last element?
						// add the ending point of the last CatchCaluse
						positions.add(new ExceptionDropPosition(cc.getStartPosition() + cc.getLength(), 1,
								ExceptionDropPosition.OUT_OF_CATCH_CLAUSE, i + 1));
					}
				}
			}
		}
	}

	/**
	 * The given node is in catch clause.
	 * 
	 * @param positions
	 * @param node
	 */
	// private static void findEndPosOfTryStatement(final
	// List<ExceptionDropPosition> positions, final ASTNode node) {
	// fLogger.info("ExceptionDropPositionFinder#findEndPosOfTryStatement");
	// TryStatement tryStatement = ASTResolving.findParentTryStatement(node);
	// if (tryStatement != null) {
	// positions.add(new ExceptionDropPosition(tryStatement.getStartPosition() +
	// tryStatement.getLength(), 1,
	// ExceptionDropPosition.KIND_CAUGHT_EXCEPTION,
	// ExceptionDropPosition.ITSELF));
	// }
	// }

	/**
	 * The given node is in "throws" at the end of method body.
	 * 
	 * @param positions
	 * @param assistant
	 */
	// private static void findEndPosOfTryStatement2(final
	// List<ExceptionDropPosition> positions,
	// final CatchExceptionAssistant assistant) {
	// fLogger.info("ExceptionDropPositionFinder#findEndPosOfTryStatement2");
	// FixBugsAnnotation annotation =
	// assistant.findFirstAnnotationInTryStatement();
	// ASTNode node = assistant.findASTNode(annotation.getPosition());
	// TryStatement tryStatement = ASTResolving.findParentTryStatement(node);
	// if (tryStatement != null) {
	// if (tryStatement.catchClauses().isEmpty()) {
	// Block block = tryStatement.getBody();
	// fLogger.info("offset: " + block.getStartPosition() + ", " +
	// block.getLength());
	// positions.add(new ExceptionDropPosition(block.getStartPosition() +
	// block.getLength(), 1,
	// ExceptionDropPosition.KIND_CAUGHT_EXCEPTION,
	// ExceptionDropPosition.ITSELF));
	// } else {
	// int size = tryStatement.catchClauses().size();
	// CatchClause lastCatchClause = (CatchClause)
	// tryStatement.catchClauses().get(size - 1);
	// positions.add(new
	// ExceptionDropPosition(lastCatchClause.getStartPosition() +
	// lastCatchClause.getLength(), 1,
	// ExceptionDropPosition.KIND_CAUGHT_EXCEPTION,
	// ExceptionDropPosition.ITSELF));
	// }
	// }
	// }

	private static int findOrderOfMultiCaughtExceptions(final SimpleName exceptionName) {
		ASTNode parent = exceptionName.getParent();
		ASTNode grandParent = parent.getParent();
		if (grandParent instanceof UnionType) {
			UnionType unionType = (UnionType) grandParent;
			int index = 0;
			for (Object obj : unionType.types()) {
				Type type = (Type) obj;
				ITypeBinding tb = type.resolveBinding();
				fLogger.fine("#getOrderInMultiCatch: " + tb.getQualifiedName());
				if (tb.getQualifiedName().contains(exceptionName.getFullyQualifiedName())) {
					return index;
				}
				index++;
			}
		}
		return -1;
	}

	private static int findOrderOfThrownException(final SimpleName exceptionName) {
		ASTNode parent = exceptionName.getParent();
		if (parent instanceof MethodDeclaration) {
			MethodDeclaration md = (MethodDeclaration) parent;
			int index = 0;
			for (Object obj : md.thrownExceptions()) {
				SimpleName ex = (SimpleName) obj;
				if (ex.equals(exceptionName)) {
					return index;
				}
				index++;
			}
		}
		return -1;
	}

	private static boolean isCaught(final SimpleName exceptionName) {
		TryStatement tryStatement = ASTResolving.findParentTryStatement(exceptionName);
		if (tryStatement != null) {
			return true;
		}
		return false;
	}

	private static boolean isMultiCaught(final SimpleName exceptionName) {
		ASTNode parent = exceptionName.getParent();
		ASTNode grandParent = parent.getParent();
		if (grandParent instanceof UnionType) {
			return true;
		}
		return false;
	}

	private static boolean isThrown(final SimpleName exceptionName) {
		ASTNode parent = exceptionName.getParent();
		if (parent instanceof MethodDeclaration) {
			return true;
		}
		return false;
	}
}
