//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.core.resolution.performance.stringBufferConcetenation;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.runtime.Assert;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.NamingConventions;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.Assignment;
import org.eclipse.jdt.core.dom.Block;
import org.eclipse.jdt.core.dom.ChildListPropertyDescriptor;
import org.eclipse.jdt.core.dom.ClassInstanceCreation;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.ExpressionStatement;
import org.eclipse.jdt.core.dom.ForStatement;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.Statement;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.VariableDeclarationExpression;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;
import org.eclipse.jdt.core.dom.rewrite.ASTRewrite;
import org.eclipse.jdt.core.dom.rewrite.ITrackedNodePosition;
import org.eclipse.jdt.core.dom.rewrite.ListRewrite;
import org.eclipse.jdt.internal.corext.dom.ASTNodes;
import org.eclipse.jdt.internal.ui.text.correction.ASTResolving;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Position;

import edu.ncsu.csc.emh.fixbugs.core.annotation.FixBugsAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.dom.ASTNodeUtil;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsContext;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsLinkedAssistant;
import edu.ncsu.csc.emh.fixbugs.core.resolution.SearchHint;
import edu.ncsu.csc.emh.fixbugs.core.util.PrintUtil;
import edu.ncsu.csc.emh.fixbugs.core.util.SuggestionUtil;
import edu.ncsu.csc.emh.fixbugs.ui.Messages;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.performance.stringBufferConcatenation.StringBufferConcatenationMode;

/**
 * @author Yoonki Song
 */
@SuppressWarnings("restriction")
public final class StringBufferConcatenationAssistant extends FixBugsLinkedAssistant {
	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(StringBufferConcatenationAssistant.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	/** */
	private String fOriginalIdentifier = null;

	/**
	 * Constructor for StringBufferConcatenationAssistant.
	 * 
	 * @param context
	 *            the invocation context.
	 * @param hint
	 *            the search hint.
	 */
	public StringBufferConcatenationAssistant(final FixBugsContext context, final SearchHint hint) {
		super(context, hint);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void init() throws BadLocationException, JavaModelException {
		super.init();
		initSeedNode(ASTNode.SIMPLE_NAME);
		initIdentifier();
		initLinkedNodesList();
		setAdjustedSelection(getFirstLinkedNodePosition());
		initAnnotations();
	}

	// ------------------ seed node --------------------//

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void initSeedNode(final int nodeType) {
		fLogger.info("StringBufferConcatenationAssistant#initSeedNode: ");
		ASTNode seedNode = findSeedNode(nodeType, getContext().getCoveringNode());
		setSeedNode(seedNode);
		fLogger.info(PrintUtil.toString("\tSeed Node Found", getSeedNode()));
	}

	/**
	 * Finds the seed node.
	 * <p>
	 * 1) Block
	 * 
	 * <pre>
	 * for (int i = 0; i &lt; 5; i++) {
	 * 	result += &quot;something&quot;;
	 * }
	 * </pre>
	 * 
	 * <p>
	 * 2) ForStatement
	 * 
	 * <pre>
	 * for (int i = 0; i &lt; 5; i++)
	 * 	result += &quot;something&quot;; // no block!!
	 * </pre>
	 * 
	 * @param nodeType
	 *            the node type.
	 * @param node
	 *            the clue AST node
	 * @return
	 */
	@Override
	public ASTNode findSeedNode(final int nodeType, final ASTNode node) {
		ASTNode seedNode = null;
		if (node instanceof Block) {
			List<Statement> statements = ((Block) node).statements();
			fLogger.info("\tstatements.size(): " + statements.size());
			for (Statement statement : statements) {
				fLogger.info("\t" + ASTNodes.asString(statement) + ", " + statement.getStartPosition());
				// the nearest statement to the current position
				if (isSelectionOverlapped(statement)) {
					seedNode = findBestGuessedVariableName(statement);
					break;
				}
			}
		} else if (node instanceof ForStatement) {
			Statement stmt = ((ForStatement) node).getBody();
			if (stmt instanceof ExpressionStatement) {
				seedNode = findBestGuessedVariableName(stmt);
			} // consider other cases!!
		} else if (node instanceof VariableDeclarationStatement) {
			seedNode = findBestGuessedVariableName(node);
		} else if (node instanceof VariableDeclarationExpression) {
			seedNode = findBestGuessedVariableName(node);
		} else if (node instanceof ExpressionStatement) {
			seedNode = findBestGuessedVariableName(node);
		} else if (node instanceof SimpleName) {
			seedNode = node;
		}
		Assert.isNotNull(seedNode);
		return seedNode;
	}

	private ASTNode findBestGuessedVariableName(final ASTNode node) {
		ASTNode guessedVariable = null;
		if (node instanceof VariableDeclarationStatement) {
			// e.g. InputStream in = new FileInputStream("name.txt");
			Type type = ((VariableDeclarationStatement) node).getType();
			String qn = type.resolveBinding().getQualifiedName();
			if (matches(qn)) {
				// first element
				Object obj = ((VariableDeclarationStatement) node).fragments().get(0);
				if (obj instanceof VariableDeclarationFragment) {
					guessedVariable = ((VariableDeclarationFragment) obj).getName();
				}
			}
		} else if (node instanceof VariableDeclarationExpression) {
			// e.g. try (InputStream in = new FileInputStream("name.txt")) {...}
			Type type = ((VariableDeclarationExpression) node).getType();
			String qn = type.resolveBinding().getQualifiedName();
			if (matches(qn)) {
				// first element
				Object obj = ((VariableDeclarationExpression) node).fragments().get(0);
				if (obj instanceof VariableDeclarationFragment) {
					guessedVariable = ((VariableDeclarationFragment) obj).getName();
				}
			}
		} else if (node instanceof ExpressionStatement) {
			// e.g. in = new FileInputStream("name.txt");
			Expression expr = ((ExpressionStatement) node).getExpression();
			if (expr instanceof Assignment) {
				Expression lhs = ((Assignment) expr).getLeftHandSide();
				String qn = lhs.resolveTypeBinding().getQualifiedName();
				if (matches(qn)) {
					guessedVariable = lhs;
				}
			}
		}
		return guessedVariable;
	}

	// ------------------ identifier --------------------//

	private void initIdentifier() {
		fOriginalIdentifier = ((SimpleName) getSeedNode()).getIdentifier();
	}

	/**
	 * 
	 * @return
	 */
	public String getOriginalIdentifier() {
		return fOriginalIdentifier;
	}

	// ------------------ annotations -------------------//

	@Override
	public void initAnnotations() {
		fLogger.info("StringBufferConcatenationAssistant#initAnnotations");
		initAnnotationNodes();
		initAnnotationTypes();
	}

	private void initAnnotationNodes() {
		List<ASTNode[]> linkedNodesList = getLinkedNodesList();
		Assert.isNotNull(linkedNodesList); // pre-condition
		List<ASTNode> annotationNodes = new ArrayList<ASTNode>();
		for (ASTNode[] nodes : linkedNodesList) {
			for (ASTNode node : nodes) { // already sorted!
				// The related part can be a Statement or an Expression!!
				ASTNode parentNode = ASTResolving.findParentStatement(node);
				if (!annotationNodes.contains(parentNode)) {
					annotationNodes.add(parentNode);
				}
			}
		}
		setAnnotationNodes(annotationNodes.toArray(new ASTNode[annotationNodes.size()]));
	}

	private void initAnnotationTypes() {
		ASTNode[] annotationNodes = getAnnotationNodes();
		Assert.isNotNull(annotationNodes); // pre-condition
		List<FixBugsAnnotation> annotations = new ArrayList<FixBugsAnnotation>();
		for (ASTNode parentNode : annotationNodes) {
			Position p = getNodePosition(parentNode);
			annotations.add(new FixBugsAnnotation(FixBugsAnnotation.RELATED, p));
		}
		setAnnotations(annotations);
	}

	// ------------------ linked nodes list --------------------//

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void initLinkedNodesList() {
		fLogger.info("StringBufferConcatenationAssistant#initLinkedNodesList");
		SimpleName name = (SimpleName) getSeedNode();
		Assert.isNotNull(name, "The name may not be null");

		// find all the same node
		ASTNode[] linkedNodes = findLinkedNodesByNode(name);

		// sort by the starting position
		Arrays.sort(linkedNodes, new Comparator<ASTNode>() {
			@Override
			public int compare(final ASTNode n1, final ASTNode n2) {
				return n1.getStartPosition() - n2.getStartPosition();
			}
		});

		List<ASTNode[]> linkedNodesList = new ArrayList<ASTNode[]>();
		linkedNodesList.add(linkedNodes);
		setLinkedNodesList(linkedNodesList);
	}

	/**
	 * Returns available variable name for the given base name.
	 * 
	 * @param baseName
	 * @return
	 */
	private String getVariableNameSuggestion(final String baseName) {
		List<String> excludedVariableNames = Arrays.asList(ASTResolving.getUsedVariableNames(getCoveringBodyDecl()));
		fLogger.fine(PrintUtil.toListString(excludedVariableNames));
		String[] variableNames = SuggestionUtil.getVariableNameSuggestions(NamingConventions.VK_LOCAL, getContext()
				.getCompilationUnit().getJavaProject(), baseName, 0, excludedVariableNames, true);
		return variableNames[0];
	}

	/**
	 * Updates ...
	 * 
	 * @param annotations
	 *            the annotations.
	 * @param varLinks
	 */
	public void update(final List<FixBugsAnnotation> annotations, List<Position> varLinks) {
		long startTime = System.currentTimeMillis();
		updateLinkedNodesList(varLinks);
		setAnnotations(annotations); // update
		fLogger.info("#update: " + (System.currentTimeMillis() - startTime) + " ms");
	}

	/**
	 * 
	 * @param newPosition
	 *            the new position
	 */
	public void updateLinkedNodesList(List<Position> varLinks) {
		fLogger.info("StringBufferConcatenationAssistant#updateLinkedNodesList: ");
		long startTime = System.currentTimeMillis();

		List<ASTNode[]> linkedNodesList = new ArrayList<ASTNode[]>();
		setLinkedNodesList(linkedNodesList);

		fLogger.info("#updateLinkedNodesList: " + (System.currentTimeMillis() - startTime) + " ms");
	}

	/**
	 * Applies the changes actually to current compilation unit.
	 * 
	 * @param suggestion
	 *            suggestion.
	 * @param trackedAnnotationMap
	 *            tracked annotation map.
	 * @param trackedExistingVariables
	 *            tracked variables.
	 * @param trackedNewVariables
	 *            tracked variables.
	 * @return the changed text
	 * 
	 * @throws BadLocationException
	 *             is thrown if one of the edits in the tree can't be executed.
	 *             The state of the document is undefined if this exception is
	 *             thrown.
	 * @throws JavaModelException
	 *             the exception is thrown if the rewrite fails.
	 */
	public String changeAST(final String suggestion, final Map<ITrackedNodePosition, FixBugsAnnotation> trackedAnnotationMap,
			final List<ITrackedNodePosition> trackedExistingVariables, List<ITrackedNodePosition> trackedNewVariables)
			throws BadLocationException, JavaModelException {
		fLogger.info("StringBufferConcatenationAssistant#changeAST");

		String changedText = null;
		if (StringBufferConcatenationMode.USE_STRING_BUFFER.equals(suggestion)) {
			changedText = changeASTInternal("StringBuffer", trackedAnnotationMap, trackedExistingVariables, trackedNewVariables);
		} else {
			changedText = changeASTInternal("StringBuilder", trackedAnnotationMap, trackedExistingVariables,
					trackedNewVariables);
		}
		return changedText;
	}

	private String changeASTInternal(final String baseName,
			final Map<ITrackedNodePosition, FixBugsAnnotation> trackedAnnotationMap,
			final List<ITrackedNodePosition> trackedExistingVariables, List<ITrackedNodePosition> trackedNewVariables)
			throws BadLocationException, JavaModelException {

		// get DOM/AST from a ICompilationUnit
		CompilationUnit root = getContext().getASTRoot();

		// creation of ASTRewrite
		AST ast = root.getAST();
		ASTRewrite rewrite = ASTRewrite.create(ast);

		// ----------------------------- Change -----------------------------//

		String curVariableName = getOriginalIdentifier();
		String newVariableName = getVariableNameSuggestion(baseName);
		ASTNode[] annotationNodes = getAnnotationNodes();

		// description of the change
		for (ASTNode node : annotationNodes) {
			Statement oldStmt = (Statement) node;
			Statement newStmt = null; // ready!

			if (ASTNodeUtil.containsPlusAssignmentOperator(oldStmt)) {
				// e.g.) s += rhs;
				Expression rhs = ASTNodeUtil.findRightHandSide(oldStmt);
				Expression newRhs = (Expression) ASTNode.copySubtree(ast, rhs);
				newStmt = createExpressionStatementWithMI(trackedNewVariables, ast, rewrite, newVariableName, "append", newRhs); //$NON-NLS-1$
				// replace existing node with new node
				if (oldStmt instanceof ExpressionStatement && (newStmt instanceof ExpressionStatement)) {
					Expression oldExpr = ((ExpressionStatement) oldStmt).getExpression();
					Expression newExpr = ((ExpressionStatement) newStmt).getExpression();
					rewrite.replace(oldExpr, newExpr, null);
					FixBugsAnnotation ma = new FixBugsAnnotation(FixBugsAnnotation.MODIFIED, Messages.modifiedAnnotation);
					trackedAnnotationMap.put(rewrite.track(newExpr), ma);
				} else {
					rewrite.replace(oldStmt, newStmt, null);
					FixBugsAnnotation ma = new FixBugsAnnotation(FixBugsAnnotation.MODIFIED, Messages.modifiedAnnotation);
					trackedAnnotationMap.put(rewrite.track(newStmt), ma);
				}
				FixBugsAnnotation ma = new FixBugsAnnotation(FixBugsAnnotation.MODIFIED, Messages.modifiedAnnotation);
				trackedAnnotationMap.put(rewrite.track(oldStmt), ma);
			} else if (ASTNodeUtil.containsInfixPlusOperator(oldStmt)) {
				// e.g.) s = s + rhs;
				Expression rho = ASTNodeUtil.findRighHandOperand(oldStmt);
				Expression newRho = (Expression) ASTNode.copySubtree(ast, rho);
				newStmt = createExpressionStatementWithMI(trackedNewVariables, ast, rewrite, newVariableName, "append", newRho); //$NON-NLS-1$

				// replace existing node with new node
				if (oldStmt instanceof ExpressionStatement && (newStmt instanceof ExpressionStatement)) {
					Expression oldExpr = ((ExpressionStatement) oldStmt).getExpression();
					Expression newExpr = ((ExpressionStatement) newStmt).getExpression();
					rewrite.replace(oldExpr, newExpr, null);
					FixBugsAnnotation ma = new FixBugsAnnotation(FixBugsAnnotation.MODIFIED, Messages.modifiedAnnotation);
					trackedAnnotationMap.put(rewrite.track(newExpr), ma);
				} else {
					rewrite.replace(oldStmt, newStmt, null);
					FixBugsAnnotation ma = new FixBugsAnnotation(FixBugsAnnotation.MODIFIED, Messages.modifiedAnnotation);
					trackedAnnotationMap.put(rewrite.track(newStmt), ma);
				}
			} else {
				FixBugsAnnotation ra = new FixBugsAnnotation(FixBugsAnnotation.RELATED, Messages.relatedAnnotation);
				trackedAnnotationMap.put(rewrite.track(oldStmt), ra);
				addLinkedPositions(trackedExistingVariables, oldStmt, fOriginalIdentifier, rewrite);
			}
		}

		Statement loopStatement = ASTNodeUtil.findParentLoopStatement(getContext().getCoveringNode());
		Block parentBlock = (Block) loopStatement.getParent();
		Assert.isNotNull(loopStatement, "Loop statement may not be null");

		// e.g) StringBuilder stringBuilder = new StringBuilder(ipAddress);
		VariableDeclarationStatement builderOrBufferStatement = (VariableDeclarationStatement) ast
				.createInstance(VariableDeclarationStatement.class);
		builderOrBufferStatement.setType(ast.newSimpleType(ast.newSimpleName(baseName)));

		VariableDeclarationFragment variableDeclarationFragment = ast.newVariableDeclarationFragment();
		SimpleName variableName = ast.newSimpleName(newVariableName);
		variableDeclarationFragment.setName(variableName);
		variableDeclarationFragment.setExtraDimensions(0);

		ClassInstanceCreation classInstanceCreation = ast.newClassInstanceCreation();
		classInstanceCreation.setExpression(null);
		classInstanceCreation.setType(ast.newSimpleType(ast.newSimpleName(baseName)));
		SimpleName sn1 = ast.newSimpleName(curVariableName);
		classInstanceCreation.arguments().add(sn1);
		classInstanceCreation.setAnonymousClassDeclaration(null);
		trackedExistingVariables.add(rewrite.track(sn1));

		variableDeclarationFragment.setInitializer(classInstanceCreation);
		builderOrBufferStatement.fragments().add(variableDeclarationFragment);

		// e.g) ipAddress = stringBuilder.toString();
		Assignment assignment = ast.newAssignment();
		SimpleName sn2 = ast.newSimpleName(curVariableName);
		assignment.setLeftHandSide(sn2);
		trackedExistingVariables.add(rewrite.track(sn2));

		MethodInvocation methodInvocation = ast.newMethodInvocation();
		SimpleName sn = ast.newSimpleName(newVariableName);
		trackedNewVariables.add(rewrite.track(sn));
		methodInvocation.setExpression(sn);
		methodInvocation.setName(ast.newSimpleName("toString")); //$NON-NLS-1$
		assignment.setRightHandSide(methodInvocation);
		ExpressionStatement newAssignmentStatement = ast.newExpressionStatement(assignment);

		ListRewrite listRewrite = rewrite.getListRewrite(parentBlock,
				(ChildListPropertyDescriptor) loopStatement.getLocationInParent());
		listRewrite.insertBefore(builderOrBufferStatement, loopStatement, null);

		// trackedVariables.add(rewrite.track(variableName)); // track!!
		trackedNewVariables.add(rewrite.track(variableName));

		FixBugsAnnotation before = new FixBugsAnnotation(FixBugsAnnotation.ADDED, Messages.addedAnnotation);
		// FixBugsAnnotation before = new
		// FixBugsAnnotation(FixBugsAnnotation.RELATED,
		// UIMessages.relatedAnnotation);
		trackedAnnotationMap.put(rewrite.track(builderOrBufferStatement), before);
		listRewrite.insertAfter(newAssignmentStatement, loopStatement, null);

		FixBugsAnnotation after = new FixBugsAnnotation(FixBugsAnnotation.ADDED, Messages.addedAnnotation);
		trackedAnnotationMap.put(rewrite.track(newAssignmentStatement), after);

		// when the unit is written on disk (and the linked mode is
		// updated), the popup should be opened!
		// System.out.println("FixBugsLinkedMode#writeOnDisk(ASTRewrite, FixBugsInvocationContext)");

		fRewrite = rewrite; // (important for quickfix)

		return rewriteMethodDecl(rewrite);
	}

	/**
	 * Creates an ExpressionStatement with MethodInvocation.
	 * 
	 * <p>
	 * Example.
	 * 
	 * <pre>
	 * builder.toString();
	 * </pre>
	 * 
	 * @param trackedVariableLinks
	 * 
	 * @param ast
	 *            the AST
	 * @param varId
	 *            the variable identifier
	 * @param methodId
	 *            the method identifier
	 * @param rhs
	 *            the right hand side.
	 * @return the new ExpressionStatement.
	 */
	private ExpressionStatement createExpressionStatementWithMI(List<ITrackedNodePosition> trackedVariableLinks, final AST ast,
			final ASTRewrite rewrite, final String varId, final String methodId, final Expression rhs) {
		return ast.newExpressionStatement(createMethodInvocation(trackedVariableLinks, ast, rewrite, varId, methodId, rhs));
	}

	/**
	 * Creates a method invocation (Expression).
	 * 
	 * <p>
	 * Example.
	 * 
	 * <pre>
	 * builder.toString()
	 * </pre>
	 * 
	 * @param trackedVariableLinks
	 * 
	 * @param ast
	 *            the AST.
	 * @param varIdentifer
	 *            the variable identifier.
	 * @param methodIdentifier
	 *            the method identifier.
	 * @param rhs
	 *            the right hand side expression.
	 * @return the new method invocation.
	 */
	private MethodInvocation createMethodInvocation(List<ITrackedNodePosition> trackedVariableLinks, final AST ast,
			final ASTRewrite rewrite, final String varIdentifer, final String methodIdentifier, final Expression rhs) {
		MethodInvocation methodInvocation = ast.newMethodInvocation();
		SimpleName varName = createSimpleName(ast, varIdentifer);
		trackedVariableLinks.add(rewrite.track(varName));
		methodInvocation.setExpression(varName);
		methodInvocation.setName(createSimpleName(ast, methodIdentifier));
		methodInvocation.arguments().add(rhs);
		return methodInvocation;
	}

	/**
	 * Creates a SimpleName.
	 * 
	 * @param trackedVariableLinks
	 * 
	 * @param ast
	 *            the AST.
	 * @param identifier
	 *            the identifier.
	 * @return new SimpleName.
	 */
	private SimpleName createSimpleName(final AST ast, final String identifier) {
		SimpleName simpleName = ast.newSimpleName(identifier);
		return simpleName;
	}

}
