//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.ui.resolutions;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.runtime.Assert;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.internal.ui.javaeditor.CompilationUnitEditor;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Position;
import org.eclipse.jface.text.source.ISourceViewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.StyledText;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;

import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsAssistant;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsStatus;
import edu.ncsu.csc.emh.fixbugs.ui.handlers.TogglePopupHandler;
import edu.ncsu.csc.emh.fixbugs.ui.services.CommandState;

/**
 * 
 * @see <a href="http://en.wikipedia.org/wiki/Strategy_pattern">Strategy
 *      pattern</a>
 * 
 * @author Yoonki Song
 */
@SuppressWarnings("restriction")
public abstract class FixBugsAnnotationMode extends FixBugsMode {
	private class EscKeyListener implements KeyListener {

		private final StyledText fTextWidget;

		public EscKeyListener() {
			final ISourceViewer viewer = fEditor.getViewer();
			fTextWidget = viewer.getTextWidget();
		}

		@Override
		public void keyPressed(final KeyEvent e) {
			if (e.keyCode == SWT.ESC) {
				fTextWidget.removeKeyListener(this);
				stop();
			}
		}

		@Override
		public void keyReleased(final KeyEvent e) {
			// do nothing
		}

		public void start() {
			fTextWidget.addKeyListener(this);
		}
	}

	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(FixBugsAnnotationMode.class.getName());

	static {
		fLogger.setLevel(Level.WARNING);
	};

	/** */
	protected static FixBugsAnnotationMode fgActiveMode;

	/**
	 * Gets active mode.
	 * 
	 * @return
	 */
	public static FixBugsAnnotationMode getActiveMode() {
		if (fgActiveMode != null) {
			ISourceViewer viewer = fgActiveMode.fEditor.getViewer();
			if (viewer != null) {
				StyledText textWidget = viewer.getTextWidget();
				if (textWidget != null && !textWidget.isDisposed()) {
					return fgActiveMode;
				}
			}
			// make sure we don't hold onto the active annotation mode if
			// anything went wrong with canceling:
			fgActiveMode = null;
		}
		return null;
	}

	/** */
	protected FixBugsAnnotationPopup fPopup;

	/** */
	protected final FixBugsAssistant fAssistant;

	/**
	 * Constructor.
	 * 
	 * @param editor
	 *            the editor.
	 * @param assistant
	 *            the assistant.
	 */
	public FixBugsAnnotationMode(final CompilationUnitEditor editor, final FixBugsAssistant assistant) {
		super(editor);
		Assert.isNotNull(editor, "Editor may not be null");
		fAssistant = assistant;
	}

	/**
	 * Applies the original.
	 */
	@Override
	public void applyOriginal() {
		fLogger.info("FixBugsAnnotationMode#applyOriginal");
		long startTime = System.currentTimeMillis();
		try {
			// 1. read changes from the original context
			String originalText = fAssistant.getCoveringBodyDeclAsString();
			writeOriginal(originalText, fAssistant.getAdjustedSelection());

			// 2. update linked modes and annotations
			fAssistant.init();
			refresh(fAssistant);
		} catch (JavaModelException e) {
			e.printStackTrace();
		} catch (BadLocationException e) {
			e.printStackTrace();
		}
		fLogger.warning("#applyOriginal: " + (System.currentTimeMillis() - startTime) + " ms");
	}

	@Override
	public void cancel() {
		fLogger.info("FixBugsAnnotationMode#cancel");
		stop();
	}

	public Position getCurrentAnnotationPosition() {
		return null; // YS implement this later
	}

	/**
	 * Returns the current context.
	 * 
	 * @param selection
	 *            the position (adjusted selection).
	 * @return the current context.
	 */
	@Override
	public abstract FixBugsAssistant getCurrentAssistant(Position selection);

	public Position getFirstAnnotationPosition() {
		return null; // YS implement this later
	}

	/**
	 * Returns the original context.
	 * 
	 * @return the original context.
	 */
	@Override
	public FixBugsAssistant getOriginalAssistant() {
		return fAssistant;
	}

	public boolean isCaretInAnnotationPosition() {
		return true; // YS implement this later
	}

	/**
	 * Refreshes annotations by given the assistant.
	 * 
	 * @param assistant
	 *            the assistant
	 * @return FixBugsStatus the FixBugs Status
	 */
	public FixBugsStatus refresh(final FixBugsAssistant assistant) {
		fLogger.info("FixBugsAnnotationMode#refresh");
		FixBugsStatus result = new FixBugsStatus();

		// update annotations
		refreshAnnotations(assistant.getAnnotations());

		// YS support animation startAnimation();

		return result;
	}

	@Override
	public void start(final boolean applyDefault) {
		fLogger.info("FixBugsAnnotationMode#start()");

		if (getActiveMode() != null) {
			// for safety; should already be handled in RenameJavaElementAction
			// YS fgActiveLinkedMode.startFullDialog();
			fLogger.warning("Annotation Mode is still activated.");
			fgActiveMode = null; // It's tricky
			return;
		}

		CommandState.toggle();
		TogglePopupHandler.showOrHide(Boolean.TRUE);

		handleUndo();

		FixBugsStatus result = null;
		try {
			FixBugsAssistant assistant = getOriginalAssistant();
			assistant.init();
			result = refresh(assistant);
		} catch (JavaModelException e) {
			e.printStackTrace();
		} catch (BadLocationException e) {
			e.printStackTrace();
		} finally {
			fgActiveMode = this;

			if (!result.hasFatalError()) {
				// System.out.println("Ok! Let's open the popup.");
				openPopup();
			} else {
				fLogger.warning(";-(");
			}
		}

		fEscKeyListener = new EscKeyListener();
		fEscKeyListener.start();
	}

	private EscKeyListener fEscKeyListener = null;

	@Override
	public void stop() {
		fLogger.info("FixBugsAnnnotationMode#stop");
		fgActiveMode = null;
		if (fPopup != null) {
			fPopup.close();
		}
		if (fEscKeyListener != null) {
			fEscKeyListener = null;
		}

		removeAnnotations();

		CommandState.toggle();
		TogglePopupHandler.showOrHide(Boolean.FALSE);
	}
}
