//
// Copyright (c) 2007-2012 Yoonki Song.
// All rights reserved. This program and the accompanying materials
// are made available under the permission of me.
//
// Contributors:
// Yoonki Song - initial API and implementation
//
package edu.ncsu.csc.emh.fixbugs.ui.resolutions;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.commands.operations.IOperationHistory;
import org.eclipse.core.commands.operations.IUndoContext;
import org.eclipse.core.commands.operations.OperationHistoryFactory;
import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.internal.ui.javaeditor.CompilationUnitEditor;
import org.eclipse.jdt.internal.ui.javaeditor.JavaEditor;
import org.eclipse.jdt.ui.JavaUI;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.ISynchronizable;
import org.eclipse.jface.text.ITextViewer;
import org.eclipse.jface.text.ITextViewerExtension6;
import org.eclipse.jface.text.IUndoManager;
import org.eclipse.jface.text.IUndoManagerExtension;
import org.eclipse.jface.text.Position;
import org.eclipse.jface.text.source.Annotation;
import org.eclipse.jface.text.source.IAnnotationModel;
import org.eclipse.jface.text.source.IAnnotationModelExtension;
import org.eclipse.jface.text.source.ISourceViewer;
import org.eclipse.ui.texteditor.IDocumentProvider;

import edu.ncsu.csc.emh.fixbugs.core.annotation.FixBugsAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsAssistant;

/**
 * 
 * @author Yoonki Song
 */
public abstract class FixBugsMode {
	/**
	 * Finds and Marks the ASTNodes that related to.
	 * 
	 * @see JavaEditor
	 */
	class AnnotationUpdaterJob extends Job {

		private final IDocument fDocument;

		private boolean fCanceled = false;

		private final List<FixBugsAnnotation> fFixBugsAnnotationList;

		public AnnotationUpdaterJob(final IDocument document, final List<FixBugsAnnotation> annotationList) {
			super("Related Statements Finder 2");
			fDocument = document;
			fFixBugsAnnotationList = annotationList;
		}

		void doCancel() {
			fCanceled = true;
			cancel();
		}

		private boolean isCanceled(final IProgressMonitor progressMonitor) {
			return fCanceled || progressMonitor.isCanceled(); // ||
																// LinkedModeModel.hasInstalledModel(fDocument);
		}

		@Override
		protected IStatus run(final IProgressMonitor monitor) {
			fLogger.info("AnnotationUpdaterJob#run: " + fFixBugsAnnotationList.size());
			if (isCanceled(monitor)) {
				return Status.CANCEL_STATUS;
			}
			ITextViewer textViewer = fEditor.getViewer();
			if (textViewer == null) {
				return Status.CANCEL_STATUS;
			}
			IDocument document = textViewer.getDocument();
			if (document == null) {
				return Status.CANCEL_STATUS;
			}

			IDocumentProvider documentProvider = fEditor.getDocumentProvider();
			if (documentProvider == null) {
				return Status.CANCEL_STATUS;
			}

			IAnnotationModel annotationModel = documentProvider.getAnnotationModel(fEditor.getEditorInput());
			if (annotationModel == null) {
				return Status.CANCEL_STATUS;
			}

			// add related statement annotations
			int length = fFixBugsAnnotationList.size();

			Map<Annotation, Position> annotationsToAdd = new HashMap<Annotation, Position>(length);

			for (FixBugsAnnotation annotation : fFixBugsAnnotationList) {
				if (isCanceled(monitor)) {
					return Status.CANCEL_STATUS;
				}
				annotationsToAdd.put(annotation, annotation.getPosition());
			}

			if (isCanceled(monitor)) {
				return Status.CANCEL_STATUS;
			}

			synchronized (getLockObject(annotationModel)) {
				if (annotationModel instanceof IAnnotationModelExtension) {
					((IAnnotationModelExtension) annotationModel).replaceAnnotations(fAnnotations, annotationsToAdd);
				} else {
					removeAnnotations();
					Iterator<Entry<Annotation, Position>> iter = annotationsToAdd.entrySet().iterator();
					while (iter.hasNext()) {
						Entry<Annotation, Position> entry = iter.next();
						annotationModel.addAnnotation(entry.getKey(), entry.getValue());
					}
				}
				fAnnotations = annotationsToAdd.keySet().toArray(new Annotation[annotationsToAdd.keySet().size()]);
			}

			return Status.OK_STATUS;
		}
	}

	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(FixBugsMode.class.getName());

	static {
		fLogger.setLevel(Level.WARNING);
	}

	/** */
	protected final CompilationUnitEditor fEditor;

	/** Annotations. */
	private Annotation[] fAnnotations = null;

	// -------------------------- Getter/Setter ----------------------------//

	/** */
	private AnnotationUpdaterJob fAnnotationUpdater;

	/** */
	private FixBugsAnnotation fFirstAnnotation;

	/**
	 * 
	 * @param editor
	 *            the editor.
	 */
	public FixBugsMode(final CompilationUnitEditor editor) {
		fEditor = editor;
	}

	/**
	 * Adds an annotation.
	 * 
	 * @param annotation
	 */
	public void addAnnotation(final FixBugsAnnotation annotation) {
		IDocumentProvider documentProvider = fEditor.getDocumentProvider();
		if (documentProvider == null) {
			return;
		}

		IAnnotationModel annotationModel = documentProvider.getAnnotationModel(fEditor.getEditorInput());
		if (annotationModel == null || fAnnotations == null) {
			return;
		}

		synchronized (getLockObject(annotationModel)) {
			annotationModel.addAnnotation(annotation, annotation.getPosition());
		}
	}

	/**
	 * Applies the default suggestion.
	 * 
	 * @return the changed context.
	 */
	public abstract FixBugsAssistant applyDefaultSuggestion();

	// ------------------------------ Operations ----------------------------//
	/**
	 * Applies the original.
	 * 
	 */
	public abstract void applyOriginal();

	/**
     *
     */
	public abstract void cancel();

	/**
	 * Returns the current context.
	 * 
	 * @return the current context.
	 */
	public abstract FixBugsAssistant getCurrentAssistant();

	/**
	 * Returns the current context.
	 * 
	 * @param selection
	 *            the position (adjusted selection).
	 * @return the current context.
	 */
	public abstract FixBugsAssistant getCurrentAssistant(Position selection);

	// ---------------------------- Annotations ----------------------------//

	public final CompilationUnitEditor getEditor() {
		return fEditor;
	}

	protected FixBugsAnnotation getFirstAnnotation() {
		return fFirstAnnotation;
	}

	/**
	 * Returns the lock object for the given annotation model.
	 * 
	 * @param annotationModel
	 *            the annotation model
	 * @return the annotation model's lock object
	 * @since 3.0
	 */
	private Object getLockObject(final IAnnotationModel annotationModel) {
		if (annotationModel instanceof ISynchronizable) {
			Object lock = ((ISynchronizable) annotationModel).getLockObject();
			if (lock != null) {
				return lock;
			}
		}
		return annotationModel;
	}

	/**
	 * Returns the current context.
	 * 
	 * @return the current context.
	 */
	public abstract FixBugsAssistant getOriginalAssistant();

	/**
	 * Updates the related statements annotations based on the current context.
	 */
	protected void handleUndo() {
		ISourceViewer viewer = fEditor.getViewer();
		if (viewer instanceof ITextViewerExtension6) {
			IUndoManager undoManager = ((ITextViewerExtension6) viewer).getUndoManager();
			if (undoManager instanceof IUndoManagerExtension) {
				IUndoManagerExtension undoManagerExtension = (IUndoManagerExtension) undoManager;
				IUndoContext undoContext = undoManagerExtension.getUndoContext();
				IOperationHistory operationHistory = OperationHistoryFactory.getOperationHistory();
			}
		}
	}

	/**
	 * Opens an option popup containing several options.
	 */
	protected abstract void openPopup();

	/**
	 * Updates the related statements annotations based on the current context.
	 * 
	 * @param annotations
	 *            the annotations.
	 */
	public final void refreshAnnotations(final List<FixBugsAnnotation> annotations) {
		Assert.isNotNull(annotations);
		IDocument document = fEditor.getViewer().getDocument();
		if (document == null) {
			return;
		}

		fFirstAnnotation = annotations.get(0);
		fAnnotationUpdater = new AnnotationUpdaterJob(document, annotations);
		fAnnotationUpdater.run(new NullProgressMonitor());
	}

	/**
	 * Removes the given annotation.
	 * 
	 * @param annotation
	 */
	public void removeAnnotation(final Annotation annotation) {
		IDocumentProvider documentProvider = fEditor.getDocumentProvider();
		if (documentProvider == null) {
			return;
		}

		IAnnotationModel annotationModel = documentProvider.getAnnotationModel(fEditor.getEditorInput());
		if (annotationModel == null || fAnnotations == null) {
			return;
		}
		// annotationModel.removeAnnotation(annotation);

		synchronized (getLockObject(annotationModel)) {
			annotationModel.removeAnnotation(annotation);
			// System.out.println("DROPZONE_REMOVED");
		}
	}

	/**
	 * Removes all annotations.
	 */
	public void removeAnnotations() {
		fLogger.info("FixBugsMode#removeAnnotations: ");
		IDocumentProvider documentProvider = fEditor.getDocumentProvider();
		if (documentProvider == null) {
			return;
		}

		IAnnotationModel annotationModel = documentProvider.getAnnotationModel(fEditor.getEditorInput());
		if (annotationModel == null || fAnnotations == null) {
			return;
		}

		synchronized (getLockObject(annotationModel)) {
			if (annotationModel instanceof IAnnotationModelExtension) {
				((IAnnotationModelExtension) annotationModel).replaceAnnotations(fAnnotations, null);
			} else {
				for (int i = 0, length = fAnnotations.length; i < length; i++) {
					annotationModel.removeAnnotation(fAnnotations[i]);
				}
			}
			fAnnotations = null;
		}

	}

	/**
	 * Starts the mode.
	 * 
	 * @param applyDefault
	 *            true if the mode starts with applying default suggestion.
	 */
	public abstract void start(boolean applyDefault);

	/**
     *
     */
	public abstract void stop();

	/**
	 * 
	 * @param originalText
	 * @param selection
	 * @throws JavaModelException
	 */
	protected final void writeOriginal(final String originalText, final Position selection) throws JavaModelException {
		fLogger.info("FixBugsMode#writeOriginal: " + selection);
		FixBugsAssistant assistant = getCurrentAssistant();
		MethodDeclaration md = assistant.getMethodUnderPosition(selection.getOffset(), selection.getLength());
		fLogger.info("#writeOriginal: " + md.getStartPosition() + ", " + md.getLength());

		// update of the compilation unit
		ICompilationUnit cu = JavaUI.getWorkingCopyManager().getWorkingCopy(getEditor().getEditorInput());
		cu.getBuffer().replace(md.getStartPosition(), md.getLength(), originalText);
		// cu.commitWorkingCopy(true, null);
	}

	/**
	 * 
	 * @param text
	 * @param selection
	 * @throws JavaModelException
	 */
	protected final void writeOriginal2Suggestion(final String text, final Position selection) throws JavaModelException {
		fLogger.info("FixBugsMode#writeOriginal2Suggestion: " + selection);
		FixBugsAssistant currentAssistant = getOriginalAssistant();
		MethodDeclaration md = currentAssistant.getMethodUnderPosition(selection.getOffset(), selection.getLength());

		// update of the compilation unit
		ICompilationUnit cu = JavaUI.getWorkingCopyManager().getWorkingCopy(getEditor().getEditorInput());
		cu.getBuffer().replace(md.getStartPosition(), md.getLength(), text);
		// cu.commitWorkingCopy(true, null);
	}

	/**
	 * current unit!!! YS the part of code area should be different based on the
	 * bug patterns/solutions
	 * 
	 * @param text
	 *            the changes
	 * @param selection
	 *            the selection
	 * @throws JavaModelException
	 * 
	 * @see //CUCorrectionProposal#createTextChange
	 */
	protected final void writeSuggestion(final String text, final Position selection) throws JavaModelException {
		fLogger.info("FixBugsMode#writeSuggestion: " + selection);
		FixBugsAssistant currentAssistant = getCurrentAssistant(selection);
		MethodDeclaration md = currentAssistant.getMethodUnderPosition(selection.getOffset(), selection.getLength());

		// update of the compilation unit
		ICompilationUnit cu = JavaUI.getWorkingCopyManager().getWorkingCopy(getEditor().getEditorInput());
		cu.getBuffer().replace(md.getStartPosition(), md.getLength(), text);
		// cu.commitWorkingCopy(true, null);
	}
}
