//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.ui.resolutions.correctness.missingArgument;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.runtime.Assert;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.rewrite.ITrackedNodePosition;
import org.eclipse.jdt.internal.ui.javaeditor.CompilationUnitEditor;
import org.eclipse.jdt.ui.JavaUI;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Position;

import edu.ncsu.csc.emh.fixbugs.core.annotation.FixBugsAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsAssistant;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsContext;
import edu.ncsu.csc.emh.fixbugs.core.resolution.correctness.missingArgument.MissingArgumentAssistant;
import edu.ncsu.csc.emh.fixbugs.core.resolution.correctness.missingArgument.VariableDropPosition;
import edu.ncsu.csc.emh.fixbugs.core.util.Pair;
import edu.ncsu.csc.emh.fixbugs.core.util.PositionTracker;
import edu.ncsu.csc.emh.fixbugs.ui.Messages;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.FixBugsAnnotationMode;

/**
 * 
 * @author Yoonki
 */
public final class MissingArgumentMode extends FixBugsAnnotationMode {
	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(MissingArgumentMode.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	/** */
	private VariableDNDManager fDNDManager = null;

	/** the presented variable for the missing argument */
	private String fPresentedVariable = null;

	/** */
	// private List<SimpleName> fSuggestibleVariables;

	/**
	 * Constructor for MissingArgumentMode.
	 * 
	 * @param editor
	 *            the editor.
	 * @param assistant
	 *            the assistant.
	 */
	public MissingArgumentMode(final CompilationUnitEditor editor, final FixBugsAssistant assistant) {
		super(editor, assistant);
	}

	// --------------------------- Assistant ------------------------------//

	@Override
	public MissingArgumentAssistant getOriginalAssistant() {
		return (MissingArgumentAssistant) fAssistant;
	}

	@Override
	public MissingArgumentAssistant getCurrentAssistant() {
		FixBugsAnnotation annotation = getFirstAnnotation();
		Position position = annotation.getPosition();
		return getCurrentAssistant(position);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public MissingArgumentAssistant getCurrentAssistant(final Position selection) {
		fLogger.info("MissingArgumentMode#getCurrentAssistant: " + selection);
		ICompilationUnit cu = JavaUI.getWorkingCopyManager().getWorkingCopy(getEditor().getEditorInput());
		FixBugsContext context = new FixBugsContext(cu, selection.getOffset(), selection.getLength());
		return new MissingArgumentAssistant(context);
	}

	// --------------------------- Popup ------------------------------//

	public MissingArgumentPopup getPopup() {
		return (MissingArgumentPopup) fPopup;
	}

	@Override
	protected void openPopup() {
		fPopup = new MissingArgumentPopup(getEditor(), this);
		fPopup.open(false);
	}

	/**
	 * There is a user operation (writing latency)!!
	 * 
	 * @param baseName
	 *            the base name
	 */
	private void reopenPopup() {
		fLogger.info("MissingArgumentMode#reopenPopup()");
		Assert.isNotNull(fPopup); // pre-condition
		fPopup = new MissingArgumentPopup(getEditor(), this);
		fPopup.open(false);
	}

	// --------------------------- Start & Stop -----------------------------//

	@Override
	public void start(boolean applyDefault) {
		super.start(applyDefault);
		VariableColorRegistry.getInstance().init();
		if (applyDefault) {
			applyDefaultSuggestion();
			if (getPopup() != null) {
				getPopup().setDefaultSuggestion(true);
			}
		}
	}

	@Override
	public void stop() {
		fLogger.info("MissingArgumentMode#stop");
		VariableColorRegistry.getInstance().dispose();
		stopDNDManager();
		super.stop();
	}

	/**
	 * Note whenever D&D is performed, shell manager is newly created!
	 * 
	 * @param assistant
	 */
	private void startDNDManager(final MissingArgumentAssistant assistant) {
		fLogger.info("MissingArgumentMode#startDNDManager()");
		Assert.isNotNull(assistant);
		fDNDManager = new VariableDNDManager(getEditor(), this, assistant);
		fDNDManager.start();
	}

	private void stopDNDManager() {
		fLogger.info("MissingArgumentMode#stopDNDManager()");
		if (fDNDManager != null) {
			fDNDManager.stop();
		}
		fDNDManager = null;
	}

	// --------------------------- Suggestions ------------------------------//

	@Override
	public void applyOriginal() {
		super.applyOriginal();
		fLogger.warning("MissingArgumentMode#applyOriginal");
		stopDNDManager();
		if (fPresentedVariable == null) { // no previous action
			getPopup().setLastChange(false);
		}
		MissingArgumentAssistant originalAssistant = getOriginalAssistant();
		startDNDManager(originalAssistant);
	}

	@Override
	public MissingArgumentAssistant applyDefaultSuggestion() {
		fLogger.info("MissingArgumentMode#applyDefaultSuggestion");
		stopDNDManager();
		MissingArgumentAssistant originalAssistant = getOriginalAssistant();
		SimpleName suggestion = originalAssistant.getSelectedVariable();
		String suggestionVariable = suggestion.getIdentifier();
		MissingArgumentAssistant currentAssistant = applySuggestion(suggestionVariable);
		startDNDManager(currentAssistant);
		return currentAssistant;
	}

	public void applyLastChange() {
		fLogger.info("MissingArgumentMode#applyLastChange");
		stopDNDManager(); // important
		String variableName = fPresentedVariable;
		MissingArgumentAssistant currentAssistant = applySuggestion(variableName);
		if (getPopup() != null) {
			getPopup().setLastChange(true);
		}
		startDNDManager(currentAssistant); // important
	}

	/**
	 * Performs user's drag and drop action.
	 * 
	 * @param variableName
	 *            the variableName
	 * @param position
	 *            the drop position
	 */
	public void performDND(final String variableName, final VariableDropPosition position) {
		fLogger.info("MissingArgumentMode#performDND: " + variableName);
		stopDNDManager(); // important
		fPresentedVariable = variableName;
		MissingArgumentAssistant currentAssistant = applySuggestion(variableName);
		if (getPopup() != null) {
			getPopup().setLastChange(true);
		}
		startDNDManager(currentAssistant); // important
	}

	/**
	 * Applies the changes actually to the current compilation unit.
	 * applyVariableArgument.
	 * 
	 * @param suggestedArgName
	 *            suggested variable name.
	 * @return
	 */
	public MissingArgumentAssistant applySuggestion(final String suggestedArgName) {
		fLogger.info("MissingArgumentMode#applySuggestion: " + suggestedArgName);
		long startTime = System.currentTimeMillis();
		MissingArgumentAssistant currentAssistant = null;
		try {
			// 1. read changes from the original context
			MissingArgumentAssistant originalAssistant = getOriginalAssistant();
			Pair<ITrackedNodePosition, FixBugsAnnotation> trackedStrLitPair = new Pair<ITrackedNodePosition, FixBugsAnnotation>();
			Pair<ITrackedNodePosition, FixBugsAnnotation> trackedPresentedArgumentPair = new Pair<ITrackedNodePosition, FixBugsAnnotation>();
			Map<ITrackedNodePosition, FixBugsAnnotation> trackedSuggestibleVariableMap = new HashMap<ITrackedNodePosition, FixBugsAnnotation>();
			String changedText = originalAssistant.changeAST(suggestedArgName, trackedStrLitPair, trackedPresentedArgumentPair,
					trackedSuggestibleVariableMap);

			// 2. write the changes to the current context
			writeSuggestion(changedText, originalAssistant.getAdjustedSelection());

			// 3. update annotations
			Position old = originalAssistant.getAdjustedSelection();
			Position missingArgumentPosition = new Position(old.getOffset(), old.getLength());
			FixBugsAnnotation maa = new FixBugsAnnotation(FixBugsAnnotation.MISSING_ARGUMENT,
					Messages.missingArgumentAnnotation, missingArgumentPosition);

			FixBugsAnnotation presentedVarAnnotation = PositionTracker.perform(trackedPresentedArgumentPair);
			List<FixBugsAnnotation> suggestibleVarAnnotations = PositionTracker.perform(trackedSuggestibleVariableMap);

			currentAssistant = getCurrentAssistant(missingArgumentPosition);
			currentAssistant.update(maa, presentedVarAnnotation, suggestibleVarAnnotations); // update
			refresh(currentAssistant);
			if (fPopup != null && fPopup.isDisposed()) {
				reopenPopup();
			}
		} catch (JavaModelException e) {
			e.printStackTrace();
		} catch (BadLocationException e) {
			e.printStackTrace();
		}
		fLogger.warning("#applySuggestion: " + (System.currentTimeMillis() - startTime) + " ms");

		return currentAssistant;
	}

}
