//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.ui.resolutions.dodgyCode.catchException;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.rewrite.ITrackedNodePosition;
import org.eclipse.jdt.internal.ui.javaeditor.CompilationUnitEditor;
import org.eclipse.jdt.ui.JavaUI;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Position;

import edu.ncsu.csc.emh.fixbugs.core.annotation.FixBugsAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsAssistant;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsContext;
import edu.ncsu.csc.emh.fixbugs.core.resolution.dodgyCode.catchException.CatchExceptionAssistant;
import edu.ncsu.csc.emh.fixbugs.core.resolution.dodgyCode.catchException.ExceptionDataModel;
import edu.ncsu.csc.emh.fixbugs.core.resolution.dodgyCode.catchException.ExceptionDropPosition;
import edu.ncsu.csc.emh.fixbugs.core.resolution.dodgyCode.catchException.ExceptionsMethodsPair;
import edu.ncsu.csc.emh.fixbugs.core.util.PositionTracker;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.FixBugsAnnotationMode;

/**
 * This class is the Controller class of MVC pattern.
 * 
 * @see ExceptionDataModel (Model)
 * @see ExceptionDNDManager (View)
 * 
 *      TODO What if there are already thrown exceptions.
 * 
 * @author Yoonki
 */
public final class CatchExceptionMode extends FixBugsAnnotationMode {

	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(CatchExceptionMode.class.getName());

	static {
		fLogger.setLevel(Level.INFO);
	}

	/** */
	public static final String MIXED = "Mixed"; //$NON-NLS-1$

	/** */
	public static final String MULTICATCH_EXCEPTIONS = "MultiCatchExceptions"; //$NON-NLS-1$

	/** */
	public static final String THROW_EXCEPTIONS = "ThrowExceptions"; //$NON-NLS-1$

	/** */
	public static final String UNICATCH_EXCEPTIONS = "UniCatchExceptions"; //$NON-NLS-1$

	/** */
	private ExceptionDataModel fCurrentDataModel = null;

	/** */
	private ExceptionDataModel fDataModel;

	/** */
	private ExceptionDNDManager fDNDManager = null;

	/**
	 * Constructor for FixBugsLinkedMode.
	 * 
	 * @param editor
	 *            the editor.
	 * @param assistant
	 *            the assistant.
	 */
	public CatchExceptionMode(final CompilationUnitEditor editor, final FixBugsAssistant assistant) {
		super(editor, assistant);
	}

	@Override
	public CatchExceptionAssistant applyDefaultSuggestion() {
		fLogger.info("CatchExceptionMode#applyDefaultSuggestion");
		CatchExceptionAssistant originalAssistant = getOriginalAssistant();
		// YS what if some exceptions may already exist
		fDataModel = new ExceptionDataModel(originalAssistant.getOriginalCaughtExceptions(),
				new ArrayList<ExceptionsMethodsPair>());
		CatchExceptionAssistant currentAssistant = applySuggestion(MULTICATCH_EXCEPTIONS, fDataModel);
		startDNDManager(currentAssistant);
		return currentAssistant;
	}

	/**
	 * 
	 */
	public void applyLastChange() {
		stopDNDManager(); // important
		ExceptionDataModel dataModel = fCurrentDataModel;
		CatchExceptionAssistant currentAssistant = applySuggestion(MULTICATCH_EXCEPTIONS, dataModel);
		if (getPopup() != null) {
			getPopup().setLastChange(true);
		}
		startDNDManager(currentAssistant); // important
	}

	@Override
	public void applyOriginal() {
		super.applyOriginal();
		if (fCurrentDataModel == null) { // no previous action
			getPopup().setLastChange(false);
		}
	}

	/**
	 * applies a suggestion.
	 * 
	 * @param dataModel
	 * @return
	 */
	public CatchExceptionAssistant applySuggestion(final String suggestion, ExceptionDataModel dataModel) {
		fLogger.info("CatchExceptionMode#applyChanges");
		long startTime = System.currentTimeMillis();
		CatchExceptionAssistant currentAssistant = null;
		try {
			// 1. read
			CatchExceptionAssistant originalAssistant = getOriginalAssistant();
			Map<ITrackedNodePosition, FixBugsAnnotation> trackedMethodNameAnnotationMap = new HashMap<ITrackedNodePosition, FixBugsAnnotation>();
			Map<ITrackedNodePosition, FixBugsAnnotation> trackedExceptionNameAnnotationMap = new HashMap<ITrackedNodePosition, FixBugsAnnotation>();
			String changedText = originalAssistant.changeAST(suggestion, dataModel, trackedMethodNameAnnotationMap,
					trackedExceptionNameAnnotationMap);

			// 2. write the changes to the current context
			writeSuggestion(changedText, originalAssistant.getAdjustedSelection());

			// 3. update annotations
			List<FixBugsAnnotation> excNameAnnotations = PositionTracker.perform(trackedExceptionNameAnnotationMap);
			List<FixBugsAnnotation> methodNameAnnotations = PositionTracker.perform(trackedMethodNameAnnotationMap);
			currentAssistant = getCurrentAssistant(excNameAnnotations.get(0).getPosition());
			currentAssistant.update(excNameAnnotations, methodNameAnnotations); // update

			refresh(currentAssistant);

			if (fPopup != null && fPopup.isDisposed()) {
				reopenPopup();
			}
		} catch (JavaModelException e) {
			e.printStackTrace();
		} catch (BadLocationException e) {
			e.printStackTrace();
		} catch (CoreException e) {
			e.printStackTrace();
		}
		fLogger.warning("#applySuggestion: " + (System.currentTimeMillis() - startTime) + " ms");
		return currentAssistant;
	}

	@Override
	public CatchExceptionAssistant getCurrentAssistant() {
		FixBugsAnnotation annotation = getFirstAnnotation();
		Position position = annotation.getPosition();
		return getCurrentAssistant(position);
	}

	@Override
	public CatchExceptionAssistant getCurrentAssistant(final Position selection) {
		ICompilationUnit cu = JavaUI.getWorkingCopyManager().getWorkingCopy(getEditor().getEditorInput());
		FixBugsContext context = new FixBugsContext(cu, selection.getOffset(), selection.getLength());
		return new CatchExceptionAssistant(context);
	}

	@Override
	public CatchExceptionAssistant getOriginalAssistant() {
		return (CatchExceptionAssistant) fAssistant;
	}

	public CatchExceptionPopup getPopup() {
		return (CatchExceptionPopup) fPopup;
	}

	private void moveFromCaughtToCaught(final String exceptionName, final ExceptionDropPosition dropPosition) {
		fLogger.info("moveFromCaughtToCaught: " + exceptionName);
		if (dropPosition.isItself()) {
			fDataModel.moveFromCaughtToCaughtNoOrder(exceptionName);
		} else {
			fDataModel.moveFromCaughtToCaughtOrder(exceptionName, dropPosition);
		}
	}

	private void moveFromCaughtToCatch(final String exceptionName, final ExceptionDropPosition dropPosition) {
		fLogger.info("moveFromCaughtToCatch: " + exceptionName);
		if (fDataModel.isMultiCaught(exceptionName)) {
			fDataModel.moveFromMultiCaughtToCatch(exceptionName, dropPosition);
		} else {
			fDataModel.moveFromCaughtToCatch(exceptionName, dropPosition);
		}
	}

	private void moveFromThrownToCaught(final String exceptionName, final ExceptionDropPosition dropPosition) {
		fLogger.info("CatchExceptionMode#moveFromThrownToCaught: " + exceptionName);
		if (dropPosition.isItself()) {
			fDataModel.moveFromThrownToCaughtNoOrder(exceptionName); // FIXME
		} else {
			fDataModel.moveFromThrownToCaughtOrder(exceptionName, dropPosition);
		}
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @see edu.ncsu.csc.emh.fixbugs.ui.resolutions.FixBugsLinkedMode#openPopup()
	 */
	@Override
	protected void openPopup() {
		fLogger.info("CatchExceptionMode#openPopup()");
		fPopup = new CatchExceptionPopup(getEditor(), this);
		fPopup.open(false);
	}

	/**
	 * Performs user's drag and drop action.
	 * 
	 * @param annotation
	 *            the source annotation being dragged and dropped.
	 * @param exceptionName
	 *            the base exception
	 * @param dropPosition
	 *            the drop position
	 */
	public void performDND(final boolean isCaught, final String exceptionName, final ExceptionDropPosition dropPosition) {
		fLogger.info("Before: \n" + fDataModel.toString());
		stopDNDManager(); // important!
		if (isCaught) {
			if (dropPosition.isInCatchClause()) {
				// from caught to caught exceptions
				moveFromCaughtToCaught(exceptionName, dropPosition);
			} else if (dropPosition.isOutOfCatchClause()) {
				// from caught to catch clause
				moveFromCaughtToCatch(exceptionName, dropPosition);
			} else if (dropPosition.isInThrownExceptions()) {
				// from caught to thrown exceptions
				fDataModel.moveFromCaughtToThrownExcpeptions(exceptionName, dropPosition);
			} else if (dropPosition.isInEmptyThrows()) {
				// from caught to throws (empty thrown exceptions)
				fDataModel.moveFromCaughtToEmptyThrows(exceptionName);
			}
		} else {
			if (dropPosition.isInCatchClause()) {
				// from thrown to caught
				moveFromThrownToCaught(exceptionName, dropPosition);
			} else if (dropPosition.isOutOfCatchClause()) {
				// from thrown to catch
				fDataModel.moveFromThrownToCatch(exceptionName, dropPosition);
			} else if (dropPosition.isInThrownExceptions()) {
				// from thrown to thrown exceptions
				fDataModel.moveFromThrownToThrownOrder(exceptionName, dropPosition);
			} else if (dropPosition.isInEmptyThrows()) {
				// from thrown to empty throws (this case will never happen).
			}
		}
		fLogger.info("After: \n" + fDataModel.toString());
		try {
			fCurrentDataModel = fDataModel.clone();
		} catch (CloneNotSupportedException e) {
			e.printStackTrace();
		} // cloneExceptionDataModel(fDataModel);
		CatchExceptionAssistant currentAssistant = applySuggestion(MIXED, fDataModel);
		if (getPopup() != null) {
			getPopup().setLastChange(true);
		}
		startDNDManager(currentAssistant); // important!
	}

	private void reopenPopup() {
		fLogger.info("CatchExceptionMode#reopenPopup()");
		Assert.isNotNull(fPopup);
		fPopup = new CatchExceptionPopup(getEditor(), this);
		fPopup.open(false);
	}

	@Override
	public void start(final boolean applyDefault) {
		fLogger.info("CatchExceptionMode#start");
		ExceptionColorRegistry.getInstance().init();
		super.start(applyDefault);

		if (applyDefault) {
			applyDefaultSuggestion();
			if (getPopup() != null) {
				getPopup().setDefaultSuggestion(true);
			}
		}
	}

	/**
	 * Note whenever D&D is performed, shell manager is newly created!
	 * 
	 * @param assistant
	 */
	private void startDNDManager(final CatchExceptionAssistant assistant) {
		fLogger.info("CatchExceptionMode#startDNDManager()");
		Assert.isNotNull(assistant);
		fDNDManager = new ExceptionDNDManager(getEditor(), this, assistant);
		fDNDManager.start();
	}

	@Override
	public void stop() {
		fLogger.info("CatchExceptionMode#stop");
		super.stop();
		ExceptionColorRegistry.getInstance().dispose();
		stopDNDManager();
	}

	private void stopDNDManager() {
		fLogger.info("CatchExceptionMode#stopDNDManager");
		if (fDNDManager != null) {
			fDNDManager.stop();
		}
		fDNDManager = null;
	}
}
