//
// Copyright (c) 2007-2012 Yoonki Song.
// All rights reserved. This program and the accompanying materials
// are made available under the permission of me.
//
// Contributors:
// Yoonki Song - initial API and implementation
//
package edu.ncsu.csc.emh.fixbugs.ui.resolutions.dodgyCode.catchException;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.jface.resource.ColorRegistry;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.widgets.Display;

/**
 * 
 * @see ColorRegistry
 * @author Yoonki Song
 */
public final class ExceptionColorRegistry {
	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(ExceptionColorRegistry.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	/** */
	private static ExceptionColorRegistry instance = null;

	private ExceptionColorRegistry() {
	}

	/**
	 * 
	 * @return
	 */
	public static synchronized ExceptionColorRegistry getInstance() {
		if (instance == null) {
			instance = new ExceptionColorRegistry();
		}
		return instance;
	}

	/** */
	private Map<String, IndexColorPair> fExceptionColorMap;

	/** */
	private List<Color> fColors = null;

	/** */
	private Color fMultiCallerColor = null;

	/**
	 * 
	 */
	public void init() {
		fLogger.info("ExceptionColorRegistry#init()");
		Display d = Display.getDefault();
		fColors = new ArrayList<Color>();
		fColors.add(new Color(d, 75, 254, 120)); // 0, 255, 0
		fColors.add(new Color(d, 51, 153, 255));
		fColors.add(new Color(d, 255, 153, 51));
		fColors.add(new Color(d, 51, 255, 153));
		fColors.add(new Color(d, 255, 0, 255));
		fColors.add(new Color(d, 0, 0, 255));
		fColors.add(new Color(d, 0, 255, 255));
		fColors.add(new Color(d, 153, 51, 255));
		fColors.add(new Color(d, 153, 255, 51));
		fColors.add(new Color(d, 255, 51, 153));
		fColors.add(new Color(d, 255, 0, 0));

		fMultiCallerColor = new Color(d, 160, 160, 160);
		fExceptionColorMap = new TreeMap<String, IndexColorPair>();
	}

	class IndexColorPair {
		public Integer fIndex;
		public Color fColor;

		IndexColorPair(final Integer index, final Color color) {
			fIndex = index;
			fColor = color;
		}
	}

	/**
	 * Registers new exception and its color.
	 * 
	 * @param exceptionName
	 *            exception name.
	 * @param hasMultiCallers
	 * @param colorIndex
	 *            color index.
	 */
	public void register(final String exceptionName, final boolean hasMultiCallers, final Integer colorIndex) {
		if (!fExceptionColorMap.containsKey(exceptionName)) {
			Color color = null;
			if (hasMultiCallers) {
				color = fMultiCallerColor;
			} else {
				color = fColors.get(colorIndex);
			}

			fExceptionColorMap.put(exceptionName, new IndexColorPair(colorIndex, color));
			fLogger.info("#register: " + colorIndex + ", color: " + color + ", Ex: " + exceptionName
					+ ", fExceptionColorMap.size: " + fExceptionColorMap.size());
		}
	}

	/**
	 * Returns the index.
	 * 
	 * @param exceptionName
	 *            exception name.
	 * @return the index.
	 */
	public int getIndex(final String exceptionName) {
		IndexColorPair pair = fExceptionColorMap.get(exceptionName);
		if (pair == null) { // Not found
			return -1;
		}
		return pair.fIndex;
	}

	/**
	 * Returns the color of given exception.
	 * 
	 * @param exceptionName
	 *            exception name.
	 * @return the color
	 */
	@SuppressWarnings("boxing")
	public Color getColor(final String exceptionName) {
		IndexColorPair pair = fExceptionColorMap.get(exceptionName);
		fLogger.info("#getColor: " + pair.fIndex + ", color: " + pair.fColor + ", Ex: " + exceptionName + ", Total: "
				+ fColors.size());
		return pair.fColor;
	}

	/**
	 * Returns the highlighted color.
	 * 
	 * @param exceptionName
	 *            the exception name.
	 * @return the Color.
	 */
	@SuppressWarnings("boxing")
	public Color getHighlightColor(final String exceptionName) {
		IndexColorPair pair = fExceptionColorMap.get(exceptionName);
		fLogger.info("#getHighlightColor: " + pair.fIndex + ", color: " + pair.fColor + ", Ex: " + exceptionName + ", Total: "
				+ fColors.size());
		return pair.fColor;
	}

	public void dispose() {
		if (fColors != null) {
			for (Color c : fColors) {
				if (!c.isDisposed()) {
					c.dispose();
				}
			}
			fColors = null;
		}
		fMultiCallerColor = null;
		fExceptionColorMap = null;
	}

	public void clear() {
		fColors.clear();
		fExceptionColorMap.clear();
	}

	public void refresh() {
		dispose();
		init();
	}
}
