//
// Copyright (c) 2012 EMH@NCSU
//
// Fast-feedback Static Analysis Tool. It works on top of FindBugs.
//
// Contributors:
//   Yoonki Song, Brittany Johnson, and Emerson Murphy-Hill
//
package edu.ncsu.csc.emh.fixbugs.ui.resolutions.performance.stringBufferConcatenation;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.eclipse.core.runtime.Assert;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.rewrite.ITrackedNodePosition;
import org.eclipse.jdt.internal.ui.javaeditor.CompilationUnitEditor;
import org.eclipse.jdt.ui.JavaUI;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Position;
import org.eclipse.text.edits.MalformedTreeException;

import edu.ncsu.csc.emh.fixbugs.core.annotation.FixBugsAnnotation;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsAssistant;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsContext;
import edu.ncsu.csc.emh.fixbugs.core.resolution.FixBugsLinkedAssistant;
import edu.ncsu.csc.emh.fixbugs.core.resolution.SearchHint;
import edu.ncsu.csc.emh.fixbugs.core.resolution.performance.stringBufferConcetenation.StringBufferConcatenationAssistant;
import edu.ncsu.csc.emh.fixbugs.core.util.PositionTracker;
import edu.ncsu.csc.emh.fixbugs.ui.resolutions.FixBugsLinkedMode;

/**
 * 
 * @author Yoonki
 */
public final class StringBufferConcatenationMode extends FixBugsLinkedMode {
	/** For logging purpose only. */
	private static Logger fLogger = Logger.getLogger(StringBufferConcatenationMode.class.getName());
	static {
		fLogger.setLevel(Level.WARNING);
	}

	/** */
	public static final String USE_STRING_BUFFER = "UseStringBuffer"; //$NON-NLS-1$

	/** */
	public static final String USE_STRING_BUILDER = "UseStringBuilder"; //$NON-NLS-1$

	/**
	 * Constructor for FixBugsLinkedMode.
	 * 
	 * @param editor
	 *            the editor.
	 * @param assistant
	 *            the assistant.
	 */
	public StringBufferConcatenationMode(final CompilationUnitEditor editor, final FixBugsLinkedAssistant assistant) {
		super(editor, assistant);
	}

	@Override
	public void start(final boolean applyDefault) {
		super.start(applyDefault);

		if (applyDefault) {
			applyDefaultSuggestion();
			if (getPopup() != null) {
				getPopup().setStringBufferSelected(true);
			}
		}
	}

	@Override
	protected void openPopup() {
		fPopup = new StringBufferConcatenationPopup(getEditor(), this);
		fPopup.open(false);
	}

	/**
	 * There is a user operation (writing latency)!!
	 * 
	 * @param baseName
	 *            the base name
	 */
	private void reopenPopup(final String baseName) {
		fLogger.info("StringBufferConcatenationMode#reopenPopup()");
		Assert.isNotNull(fPopup);
		fPopup = new StringBufferConcatenationPopup(getEditor(), this);
		fPopup.open(false);
		if ("StringBuffer".equals(baseName)) { //$NON-NLS-1$
			((StringBufferConcatenationPopup) fPopup).setStringBufferSelected(true);
		} else {
			((StringBufferConcatenationPopup) fPopup).setStringBuilderSelected(true);
		}
	}

	@Override
	public StringBufferConcatenationPopup getPopup() {
		return (StringBufferConcatenationPopup) fPopup;
	}

	@Override
	public StringBufferConcatenationAssistant getOriginalAssistant() {
		return (StringBufferConcatenationAssistant) fAssistant;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public StringBufferConcatenationAssistant getCurrentAssistant(final Position selection) {
		fLogger.info("StringBufferConcatenationMode#getCurrentAssistant: " + selection);
		ICompilationUnit cu = JavaUI.getWorkingCopyManager().getWorkingCopy(getEditor().getEditorInput());
		FixBugsContext context = new FixBugsContext(cu, selection.getOffset(), selection.getLength());
		SearchHint hint = new SearchHint();
		hint.add("String");
		return new StringBufferConcatenationAssistant(context, hint);
	}

	@Override
	public StringBufferConcatenationAssistant getCurrentAssistant() {
		Position p = getFirstLinkedPosition();
		return getCurrentAssistant(p);
	}

	@Override
	public FixBugsAssistant applyDefaultSuggestion() {
		fLogger.info("StringBufferConcatenationMode#applyDefaultSuggestion");
		return applySuggestion(USE_STRING_BUFFER, true);
	}

	/**
	 * 
	 * @param suggestion
	 *            string representation of suggestion.
	 * @param fromOriginal
	 * @return
	 */
	public StringBufferConcatenationAssistant applySuggestion(final String suggestion, final boolean fromOriginal) {
		fLogger.info("StringBufferConcatenationMode#applySuggestion: " + suggestion + ", " + fromOriginal);
		long startTime = System.currentTimeMillis();
		StringBufferConcatenationAssistant currentAssistant = null;
		try {
			// 1. read
			StringBufferConcatenationAssistant originalAssistant = getOriginalAssistant(); // original
																							// context
			Map<ITrackedNodePosition, FixBugsAnnotation> trackedAnnotationMap = new HashMap<ITrackedNodePosition, FixBugsAnnotation>();
			List<ITrackedNodePosition> trackedExistingVariableLinks = new ArrayList<ITrackedNodePosition>();
			List<ITrackedNodePosition> trackedNewVariableLinks = new ArrayList<ITrackedNodePosition>();

			String changedText = originalAssistant.changeAST(suggestion, trackedAnnotationMap, trackedExistingVariableLinks,
					trackedNewVariableLinks);

			// 2. write on current unit
			writeSuggestion(changedText, originalAssistant.getAdjustedSelection()); // here#1!!

			// 3. update linked mode and annotations
			Position firstLinkedNode = PositionTracker.perform(trackedExistingVariableLinks.get(0));
			currentAssistant = getCurrentAssistant(firstLinkedNode); // here#2!!

			List<Position> existingVarLinks = PositionTracker.perform(trackedExistingVariableLinks);
			List<Position> newVarLinks = PositionTracker.perform(trackedNewVariableLinks);

			currentAssistant.update(PositionTracker.perform(trackedAnnotationMap), newVarLinks); // update

			List<List<Position>> linkedNodesList = new ArrayList<List<Position>>();
			linkedNodesList.add(existingVarLinks);
			linkedNodesList.add(newVarLinks);
			refresh(currentAssistant, linkedNodesList);

			// YS Temporary fix!
			if (fPopup != null && fPopup.isDisposed()) {
				reopenPopup(suggestion);
			}
		} catch (JavaModelException e) {
			e.printStackTrace();
		} catch (MalformedTreeException e) {
			e.printStackTrace();
		} catch (BadLocationException e) {
			e.printStackTrace();
		}

		fLogger.warning("#applySuggestion: " + (System.currentTimeMillis() - startTime) + " ms");
		return currentAssistant;
	}
}
